Package: surviv
Type: Package
Title: Flexible Instrumental Variable Methods for Survival Analysis
Version: 0.1.3
Authors@R: c(
    person("Haobin", "Chen", email = "tony.haobin.chen@alumni.emory.edu", role = c("aut", "cre")),
    person("James", "O'Malley", email = "James.OMalley@dartmouth.edu", role = "aut"),
    person("Pablo", "Martinez-Camblor", role = "aut"), 
    person("Todd", "MacKenzie", role = "aut")
    )
Description: Instrumental variable analysis methods for causal inference on survival data based on the Cox model allowing for various treatment and effect types, including orthogonality method-of-moments instrumental variable estimation for the Cox model, two-stage residual inclusion Cox estimation with frailty, sequential trial emulation, sequential Cox analyses, and sequential two-stage residual inclusion Cox analyses. Methodological background includes MacKenzie et al. (2014) <doi:10.1007/s10742-014-0117-x>, Martinez-Camblor et al. (2019) <doi:10.1093/biostatistics/kxx062>, Martinez-Camblor et al. (2019) <doi:10.1111/rssc.12341>, Gran et al. (2010) <doi:10.1002/sim.4048>, and Keogh et al. (2023) <doi:10.1002/sim.9718>.
License: MIT + file LICENSE
Depends: R (>= 2.10)
Imports: MASS, survival (>= 3.4-0), rootSolve
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.2.3
URL: https://github.com/tonyhbc/surviv
BugReports: https://github.com/tonyhbc/surviv/issues
NeedsCompilation: no
Packaged: 2026-04-22 16:52:05 UTC; tonyhchen
Author: Haobin Chen [aut, cre],
  James O'Malley [aut],
  Pablo Martinez-Camblor [aut],
  Todd MacKenzie [aut]
Maintainer: Haobin Chen <tony.haobin.chen@alumni.emory.edu>
Repository: CRAN
Date/Publication: 2026-04-24 18:30:02 UTC
Built: R 4.4.3; ; 2026-04-24 19:47:29 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,coxivomom)
S3method(print,coxivseq)
S3method(print,coxivtsrif)
S3method(print,seqcox)
S3method(print,seqem)
export(coxiv_omom)
export(coxiv_seq)
export(coxiv_tsrif)
export(seqcox)
export(seqem)
import(MASS)
import(stats)
import(survival)

#' Sequential Cox analysis via sequential trials emulation
#'
#' @description
#' Fit a composite Cox model to a stacked sequential trial emulation object
#' produced by [seqem()]. The analysis follows the sequential Cox approach of
#' Gran et al. by using stabilized inverse probability of artificial censoring
#' weights (IPACW) to adjust for the informative artificial censoring induced by
#' restricting each emulated trial to individuals who continue to follow their
#' treatment assignment at the trial start.
#'
#' @param formula a survival formula for the composite Cox model, typically of
#'   the form `Surv(start.new, stop.new, event) ~ trt.base + L.base + ...`.
#'   The *first* `.base` term is treated as the trial-baseline treatment variable.
#' @param data a `seqem` object returned by [seqem()]. A raw data frame is not
#'   accepted.
#' @param id a string name of the subject identifier column in the sequential trials data.
#' @param trial_id a string name of the trial index column. Defaults to `"trial"`.
#' @param stratify logical; if `TRUE`, the composite Cox model is stratified by
#'   `trial_id` so each emulated trial has its own baseline hazard.
#' @param se_type standard error estimator: one of `"robust"`, `"jackknife"`,
#'   `"bootstrap"`, or `"none"`.
#' @param B number of bootstrap replicates when `se_type = "bootstrap"`.
#' @param seed optional random seed for the bootstrap.
#' @param ipacw_trunc optional truncation quantile for cumulative stabilized
#'   IPACW. Defaults to `0.95`. Set to `NULL` to disable truncation.
#' @param ipacw_den_covs optional character vector of underlying covariate names
#'   to use in the denominator IPACW model through their lead versions (for
#'   example `c("endoleak")` uses `endoleak.lead1`). If `NULL`, the underlying
#'   variables corresponding to non-treatment `.base` terms in `formula` are
#'   used.
#' @param ipacw_num_covs optional character vector of baseline `.base`
#'   covariates to use in the numerator stabilizer model. If `NULL`, an
#'   intercept-only numerator model is used.
#' @param by_trial logical; if `TRUE` (default), fit trial-specific Cox models
#'   and return them in `$fit_by_trial`.
#' @param ... additional arguments passed to [survival::coxph()].
#'
#' @return A `seqcox` object with components including:
#' * `call` : the matched call.
#' * `formula` : the composite Cox formula actually fitted.
#' * `fit` : a composite [survival::coxph()] fit.
#' * `coef` : a `data.frame` of estimated coefficient summaries with columns
#'          `logHR`, `HR`, `se`, `Z`, and `p.val`.
#' * `conf_int` : a `data.frame` of 95% confidence limits with columns
#'         `2.5%` and `97.5%`.
#' * `n_ids`, `n_trials`, `n_event`: the analysis size summaries
#'         for sample size, number of emulated trials aggregated, and number
#'         of events.
#' * `fit_by_trial` : a list of trial-specific Cox fits when
#'         `by_trial = TRUE`.
#' * `data_ste` : a `data.frame` of stacked analysis dataset actually used in
#'         the composite fit, including the computed IPACW column.
#' * `jackknife`, `bootstrap`: a vector of jackknife or bootstrap resampled
#'         estimates if `se_type` set to `jackknife` or `boostrap`.
#'
#' @references
#' Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B,
#' Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the
#' causal effect of treatment in the presence of time-dependent confounding
#' applied to data from the Swiss HIV Cohort Study. \emph{Statistics in
#' Medicine}. 2010;29(26):2757-2768.
#'
#' Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in
#' survival analysis using longitudinal observational data: Sequential trials
#' and marginal structural models. \emph{Statistics in Medicine}.
#' 2023;42(13):2191-2225.
#'
#' @examples
#'   data("vascular", package = "surviv")
#'   vasc_seqem = seqem(data = vascular, start = "time.start",
#'                      stop = "time.stop", event = "event", id = "id",
#'                      tvtrt = "reint", covs = c("diameter", "endoleak"),
#'                      coarsen = "ceiling", cbin_width = 2)
#'   fit_seqcox <- seqcox(
#'     formula = survival::Surv(start.new, stop.new, event) ~
#'       reint.base + diameter.base + endoleak.base,
#'     data = vasc_seqem,
#'     id = "id",
#'     se_type = "robust"
#'   )
#'   print(fit_seqcox)
#'
#' @import survival
#' @export
seqcox <- function(formula,
                   data,
                   id,
                   trial_id = "trial",
                   stratify = FALSE,
                   se_type = c("robust", "jackknife", "bootstrap", "none"),
                   B = 50,
                   seed = NULL,
                   ipacw_trunc = 0.95,
                   ipacw_den_covs = NULL,
                   ipacw_num_covs = NULL,
                   by_trial = TRUE,
                   ...) {

  cl <- match.call()
  se_type <- match.arg(se_type)

  if (!inherits(data, "seqem")) {
    stop("`data` must be a `seqem` object produced by `seqem()`.")
  }

  dat_orig <- data$data_seqorig
  if (!is.data.frame(dat_orig)) {
    stop("`data` must contain `data_seqorig` from `seqem()`.")
  }

  needed_core <- c(id, trial_id, "rownum")
  if (!all(needed_core %in% names(dat_orig))) {
    stop("Sequential trials data are missing one or more required columns: ",
         paste(setdiff(needed_core, names(dat_orig)), collapse = ", "))
  }

  if (!inherits(formula, "formula")) {
    stop("`formula` must be a formula.")
  }

  tt <- stats::terms(formula)
  rhs_terms <- attr(tt, "term.labels")
  base_terms <- rhs_terms[grepl("\\.base$", rhs_terms)]

  if (length(base_terms) == 0L) {
    stop("The RHS of `formula` must contain at least one `.base` term.")
  }

  trt_base_var <- base_terms[1L]
  strip_base <- function(x) sub("\\.base$", "", x)
  trt_var <- strip_base(trt_base_var)

  if (!all(c(trt_var, trt_base_var) %in% names(dat_orig))) {
    stop("Could not find treatment columns `", trt_var, "` and/or `",
         trt_base_var, "` in `data_seqorig`.")
  }

  # Baseline covariates appearing in the outcome model, excluding treatment.
  baseline_cov_base <- setdiff(base_terms, trt_base_var)

  # Denominator model should use time-varying lead versions.
  if (is.null(ipacw_den_covs)) {
    tv_covs <- unique(strip_base(baseline_cov_base))
  } else {
    tv_covs <- ipacw_den_covs
  }

  dat <- .compute_ipacw_seq(
    dat_seq_orig = dat_orig,
    id_var = id,
    trial_var = trial_id,
    trt_var = trt_var,
    trt_base_var = trt_base_var,
    rownum_var = "rownum",
    tv_cov_lead = tv_covs,
    baseline_covs_base = ipacw_num_covs,
    trunc_prob = ipacw_trunc
  )

  if (!"ipacw.stab.cum.trunc" %in% names(dat)) {
    stop("IPACW computation failed to produce `ipacw.stab.cum.trunc`.")
  }

  dat$.seqcox_wt <- as.numeric(dat$ipacw.stab.cum.trunc)

  add_strata <- function(formula, trial_id) {
    f_chr <- paste(deparse(formula), collapse = " ")
    if (grepl("strata(", f_chr, fixed = TRUE)) return(formula)
    parts <- strsplit(f_chr, "~", fixed = TRUE)[[1L]]
    lhs <- trimws(parts[1L])
    rhs <- trimws(parts[2L])
    rhs_new <- if (rhs == "" || rhs == "1") {
      sprintf("strata(%s)", trial_id)
    } else {
      paste(rhs, sprintf("strata(%s)", trial_id), sep = " + ")
    }
    stats::as.formula(paste(lhs, "~", rhs_new), env = environment(formula))
  }

  add_cluster <- function(formula, id_var) {
    f_chr <- paste(deparse(formula), collapse = " ")
    if (grepl("cluster(", f_chr, fixed = TRUE)) return(formula)
    parts <- strsplit(f_chr, "~", fixed = TRUE)[[1L]]
    lhs <- trimws(parts[1L])
    rhs <- trimws(parts[2L])
    rhs_new <- if (rhs == "" || rhs == "1") {
      sprintf("cluster(%s)", id_var)
    } else {
      paste(rhs, sprintf("cluster(%s)", id_var), sep = " + ")
    }
    stats::as.formula(paste(lhs, "~", rhs_new), env = environment(formula))
  }

  form_fit <- formula
  if (isTRUE(stratify)) {
    form_fit <- add_strata(form_fit, trial_id)
  }
  form_trial <- formula

  fit_main <- survival::coxph(
    formula = form_fit,
    data = dat,
    weights = .seqcox_wt,
    ...
  )

  coef_hat <- stats::coef(fit_main)
  if (is.null(coef_hat) || length(coef_hat) == 0L) {
    stop("No coefficients were estimated by the composite Cox model.")
  }

  var_hat <- fit_main$var
  fit_used <- fit_main
  jackknife_est <- NULL
  bootstrap_est <- NULL

  if (se_type == "robust") {
    form_rob <- add_cluster(form_fit, id)
    fit_rob <- survival::coxph(
      formula = form_rob,
      data = dat,
      weights = .seqcox_wt,
      robust = TRUE,
      ...
    )
    coef_hat <- stats::coef(fit_rob)
    var_hat <- fit_rob$var
    fit_used <- fit_rob
  } else if (se_type == "jackknife") {
    id_vec <- dat[[id]]
    ids_unique <- unique(id_vec)
    p <- length(coef_hat)
    jackknife_est <- matrix(NA_real_, nrow = length(ids_unique), ncol = p)
    colnames(jackknife_est) <- names(coef_hat)
    rows_by_id <- split(seq_len(nrow(dat)), id_vec)

    for (j in seq_along(ids_unique)) {
      keep_rows <- unlist(rows_by_id[setdiff(names(rows_by_id), ids_unique[j])],
                          use.names = FALSE)
      dat_j <- dat[keep_rows, , drop = FALSE]
      fit_j <- try(
        survival::coxph(form_fit, data = dat_j, weights = .seqcox_wt, ...),
        silent = TRUE
      )
      if (!inherits(fit_j, "try-error")) {
        jackknife_est[j, ] <- stats::coef(fit_j)
      }
    }

    ok <- stats::complete.cases(jackknife_est)
    if (sum(ok) > 1L) {
      jk <- jackknife_est[ok, , drop = FALSE]
      m <- nrow(jk)
      jk_centered <- sweep(jk, 2L, coef_hat, FUN = "-")
      var_hat <- (m - 1) / m * crossprod(jk_centered)
    } else {
      warning("Too few successful jackknife fits; using model-based variance.")
    }
  } else if (se_type == "bootstrap") {
    if (!is.null(seed)) set.seed(seed)
    id_vec <- dat[[id]]
    ids_unique <- unique(id_vec)
    p <- length(coef_hat)
    bootstrap_est <- matrix(NA_real_, nrow = B, ncol = p)
    colnames(bootstrap_est) <- names(coef_hat)
    rows_by_id <- split(seq_len(nrow(dat)), id_vec)

    for (b in seq_len(B)) {
      ids_b <- sample(ids_unique, size = length(ids_unique), replace = TRUE)
      idx_b <- unlist(rows_by_id[match(ids_b, ids_unique)], use.names = FALSE)
      dat_b <- dat[idx_b, , drop = FALSE]
      fit_b <- try(
        survival::coxph(form_fit, data = dat_b, weights = .seqcox_wt, ...),
        silent = TRUE
      )
      if (!inherits(fit_b, "try-error")) {
        bootstrap_est[b, ] <- stats::coef(fit_b)
      }
    }

    ok <- stats::complete.cases(bootstrap_est)
    if (sum(ok) > 1L) {
      var_hat <- stats::cov(bootstrap_est[ok, , drop = FALSE])
    } else {
      warning("Too few successful bootstrap fits; using model-based variance.")
    }
  }

  se_hat <- sqrt(diag(var_hat))
  z_hat <- coef_hat / se_hat
  p_hat <- 2 * stats::pnorm(abs(z_hat), lower.tail = FALSE)

  conf_low <- coef_hat - stats::qnorm(0.975) * se_hat
  conf_high <- coef_hat + stats::qnorm(0.975) * se_hat

  coef_df <- data.frame(
    logHR = as.numeric(coef_hat),
    HR = exp(as.numeric(coef_hat)),
    se = as.numeric(se_hat),
    Z = as.numeric(z_hat),
    p.val = as.numeric(p_hat),
    row.names = names(coef_hat),
    check.names = FALSE
  )

  conf_int <- data.frame(
    "2.5%" = as.numeric(conf_low),
    "97.5%" = as.numeric(conf_high),
    row.names = names(coef_hat),
    check.names = FALSE
  )

  fit_by_trial <- vector("list", 0L)
  if (isTRUE(by_trial)) {
    trial_vals <- sort(unique(dat[[trial_id]]))
    fit_by_trial <- vector("list", length(trial_vals))
    names(fit_by_trial) <- paste0("fit_et", seq_along(trial_vals))

    for (k in seq_along(trial_vals)) {
      dat_k <- dat[dat[[trial_id]] == trial_vals[k], , drop = FALSE]
      fit_k <- try(
        survival::coxph(form_trial, data = dat_k, weights = .seqcox_wt, ...),
        silent = TRUE
      )
      fit_by_trial[[k]] <- if (inherits(fit_k, "try-error")) NULL else fit_k
    }
  }

  out <- list(
    call = cl,
    formula = form_fit,
    fit = fit_used,
    coef = coef_df,
    conf_int = conf_int,
    n = fit_used$n,
    n_event = fit_used$nevent,
    n_ids = length(unique(dat[[id]])),
    n_trials = length(unique(dat[[trial_id]])),
    trial_id = trial_id,
    jackknife = jackknife_est,
    bootstrap = bootstrap_est,
    fit_by_trial = fit_by_trial,
    data_ste = dat
  )

  class(out) <- "seqcox"
  out
}


# Internal helper: stabilized IPACW for sequential trials data
.compute_ipacw_seq <- function(dat_seq_orig,
                               id_var,
                               trial_var,
                               trt_var,
                               trt_base_var,
                               rownum_var = "rownum",
                               tv_cov_lead = NULL,
                               baseline_covs_base = NULL,
                               trunc_prob = 0.95) {

  stopifnot(id_var %in% names(dat_seq_orig),
            trial_var %in% names(dat_seq_orig),
            trt_var %in% names(dat_seq_orig),
            trt_base_var %in% names(dat_seq_orig),
            rownum_var %in% names(dat_seq_orig))

  ord <- order(dat_seq_orig[[id_var]],
               dat_seq_orig[[trial_var]],
               dat_seq_orig[[rownum_var]])
  dat_seq_orig <- dat_seq_orig[ord, , drop = FALSE]

  g <- interaction(dat_seq_orig[[id_var]], dat_seq_orig[[trial_var]], drop = TRUE)

  dat_seq_orig$A.lead1 <- ave(dat_seq_orig[[trt_var]], g,
                              FUN = function(x) c(x[-1], NA))

  dat_seq_orig$Anext.equal.to.baseline <- ifelse(
    is.na(dat_seq_orig$A.lead1),
    NA_integer_,
    ifelse(dat_seq_orig$A.lead1 == dat_seq_orig[[trt_base_var]], 1L, 0L)
  )

  lead_cov_names <- character(0L)
  if (!is.null(tv_cov_lead) && length(tv_cov_lead) > 0L) {
    for (v in tv_cov_lead) {
      if (!v %in% names(dat_seq_orig)) next
      newname <- paste0(v, ".lead1")
      dat_seq_orig[[newname]] <- ave(dat_seq_orig[[v]], g,
                                     FUN = function(x) c(x[-1], NA))
      lead_cov_names <- c(lead_cov_names, newname)
    }
  }

  keep_idx <- dat_seq_orig[[trt_var]] == dat_seq_orig[[trt_base_var]]
  dat_seq <- dat_seq_orig[keep_idx, , drop = FALSE]

  dat_seq$wt.denom <- 1
  dat_seq$wt.num <- 1

  mask_est <- (dat_seq[[trt_base_var]] == 0) & !is.na(dat_seq$Anext.equal.to.baseline)

  if (any(mask_est)) {
    est_data <- dat_seq[mask_est, , drop = FALSE]

    # Denominator uses lead versions of time-varying covariates when available.
    if (length(lead_cov_names) > 0L) {
      form_denom <- stats::as.formula(
        paste("Anext.equal.to.baseline ~", paste(lead_cov_names, collapse = " + "))
      )
    } else {
      form_denom <- Anext.equal.to.baseline ~ 1
    }

    wt.mod.denom <- stats::glm(form_denom,
                               family = stats::binomial(),
                               data = est_data)

    dat_seq$wt.denom[mask_est] <- stats::predict(
      wt.mod.denom,
      newdata = est_data,
      type = "response"
    )

    # By default use an intercept-only numerator stabilizer.
    base_covs_in_data <- NULL
    if (!is.null(baseline_covs_base)) {
      base_covs_in_data <- baseline_covs_base[baseline_covs_base %in% names(est_data)]
    }

    if (is.null(base_covs_in_data) || length(base_covs_in_data) == 0L) {
      form_num <- Anext.equal.to.baseline ~ 1
    } else {
      form_num <- stats::as.formula(
        paste("Anext.equal.to.baseline ~", paste(base_covs_in_data, collapse = " + "))
      )
    }

    wt.mod.num <- stats::glm(form_num,
                             family = stats::binomial(),
                             data = est_data)

    dat_seq$wt.num[mask_est] <- stats::predict(
      wt.mod.num,
      newdata = est_data,
      type = "response"
    )
  }

  dat_seq$wt.denom <- ifelse(dat_seq[[trt_base_var]] == 1, 1, dat_seq$wt.denom)
  dat_seq$wt.num <- ifelse(dat_seq[[trt_base_var]] == 1, 1, dat_seq$wt.num)

  dat_seq$ipacw.stab <- dat_seq$wt.num / dat_seq$wt.denom

  g2 <- interaction(dat_seq[[id_var]], dat_seq[[trial_var]], drop = TRUE)
  dat_seq$ipacw.stab.lag <- ave(dat_seq$ipacw.stab, g2,
                                FUN = function(x) c(1, x[-length(x)]))
  dat_seq$ipacw.stab.cum <- ave(dat_seq$ipacw.stab.lag, g2, FUN = cumprod)

  if (!is.null(trunc_prob)) {
    pct <- stats::quantile(dat_seq$ipacw.stab.cum,
                           probs = trunc_prob,
                           na.rm = TRUE)
    dat_seq$ipacw.stab.cum.trunc <- pmin(dat_seq$ipacw.stab.cum, pct)
  } else {
    dat_seq$ipacw.stab.cum.trunc <- dat_seq$ipacw.stab.cum
  }

  dat_seq
}

#' Print method for seqem objects
#'
#' @param x An object of class `seqem` returned by [seqem()].
#' @param ... Ignored.
#'
#' @return The input object `x`, invisibly. Called for its side effect of printing a summary to the console.
#'
#' @export
print.seqem <- function(x, ...) {
  if (!inherits(x, "seqem")) {
    return(NextMethod())
  }

  if (!is.null(x$data_seqem) && is.data.frame(x$data_seqem)) {
    cat(.seqem_summary_text(x$data_seqem, trial_summary = x$trial_summary))
  } else if (!is.null(x$summary) && is.character(x$summary)) {
    cat(x$summary)
  } else {
    cat("Object of class 'seqem' with no summary available.\n")
  }

  invisible(x)
}

#' Sequential 2SRI Cox analysis via sequential trials emulation
#'
#' Fits a sequential two-stage residual inclusion (2SRI) Cox
#' model for a time-varying treatment with a baseline instrumental variable,
#' using a stacked sequential trial emulation created by [seqem()].
#'
#' The analysis proceeds trial by trial. Within each emulated trial, a
#' first-stage logistic regression is fit for the trial-baseline treatment
#' `{tvtrt}.base` on the baseline IV and baseline covariates from
#' `trtformula`. A trial-specific control-function residual `R.base` is then
#' constructed and carried across all rows within that subject-trial trajectory.
#'
#' The stacked data are then artificially censored when observed treatment
#' deviates from the trial-baseline treatment status, and stabilised inverse
#' probability of artificial censoring weights (IPACW) are estimated. These
#' weights address the informative censoring induced by the sequential trial
#' emulation itself. This function does not estimate additional weights for
#' naturally occurring right censoring.
#'
#' Finally, a weighted composite Cox model is fit on the artificially censored,
#' weighted stacked data. The returned coefficient summary is organised in the
#' same style as [seqcox()], [coxiv_omom()], and [coxiv_tsrif()].
#'
#' @param formula a `formula` for the composite second-stage Cox model,
#'   typically of the form
#'   `Surv(start.new, stop.new, event) ~ tvtrt.base + x1.base + x2.base`.
#'   The right-hand side must include `paste0(tvtrt, ".base")`. The control
#'   function term `R.base` is added automatically if absent.
#' @param trtformula a `formula` for the trial-specific first-stage treatment
#'   model. Its left-hand side must equal `paste0(tvtrt, ".base")`, and its
#'   right-hand side must include `paste0(iv, ".base")`.
#' @param data a `seqem` object returned by [seqem()].
#' @param id a string name for the subject identifier variable.
#' @param tvtrt a string name for the original time-varying treatment
#'   variable.
#' @param iv a string name for the original baseline instrumental
#'   variable.
#' @param stratify logical; if `TRUE`, the composite Cox model is stratified by
#'   emulated trial so that each trial has its own baseline hazard.
#' @param se_type standard error estimator: one of `"robust"`, `"jackknife"`,
#'   `"bootstrap"`, or `"none"`.
#' @param B the number of bootstrap replicates when `se_type = "bootstrap"`.
#' @param seed optional random seed for bootstrap resampling.
#' @param ipacw_den_covs optional string vector of underlying covariate names
#'   to use in the denominator IPACW model through their lead versions (for
#'   example `c("endoleak")` uses `endoleak.lead1`). If `NULL`, the underlying
#'   variables corresponding to non-treatment `.base` terms in `formula` and
#'   `trtformula` are used whenever the original variables are available in the
#'   stacked data.
#' @param ipacw_num_covs optional string vector of baseline `.base`
#'   covariates to use in the numerator stabiliser model. If `NULL`, an
#'   intercept-only numerator model is used.
#' @param by_trial logical; if `TRUE`, fit and return separate second-stage Cox
#'   models within each emulated trial.
#' @param ... additional arguments passed to [survival::coxph()].
#'
#' @return A `coxivseq` object with components including:
#' \itemize{
#'   \item `call` : the matched call.
#'   \item `formula` : the composite second-stage Cox formula actually fitted.
#'   \item `trtformula` : the sequential first-stage treatment models formula.
#'   \item `fit` : the fitted composite [survival::coxph()] model.
#'   \item `coef` : a `data.frame` with columns `logHR`, `HR`, `se`, `Z`, `p.val`.
#'   \item `conf_int` : a `data.frame` of 95% confidence limits with columns
#'         `2.5%` and `97.5%` for estimated log hazard ratios.
#'   \item `vcov_mat` : variance-covariance matrix corresponding to the
#'         selected `se_type`.
#'   \item `trtfit_by_trial` : list of first-stage treatment models.
#'   \item `fit_by_trial` : list of second-stage trial-specific Cox fits when
#'         `by_trial = TRUE`.
#'   \item `data_ste` : final stacked, artificially censored analysis dataset,
#'         including `R.base` and the computed IPACW column.
#'   \item `jackknife`, `bootstrap`: a `matrix` of jackknife or bootstrap resampled
#'         estimates if `se_type` set to `jackknife` or `boostrap`.
#'   \item `tvtrt`, `iv`, `n`, `n_ids`, `n_trials`, `n_event`, `stratify`,
#'         `se_type`, `B`, `by_trial`: miscellaneous string tags for model print
#'         summary.
#' }
#'
#' @references
#' Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B,
#' Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the
#' causal effect of treatment in the presence of time-dependent confounding
#' applied to data from the Swiss HIV Cohort Study. \emph{Statistics in
#' Medicine}. 2010;29(26):2757-2768.
#'
#' Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in
#' survival analysis using longitudinal observational data: Sequential trials
#' and marginal structural models. \emph{Statistics in Medicine}.
#' 2023;42(13):2191-2225.
#'
#' @examples
#' data("vascular", package = "surviv")
#'
#' vasc_seqem <- seqem(
#'   data = vascular,
#'   start = "time.start",
#'   stop = "time.stop",
#'   event = "event",
#'   id = "id",
#'   tvtrt = "reint",
#'   covs = c("iv", "diameter", "endoleak"),
#'   coarsen = "ceiling",
#'   cbin_width = 2
#' )
#'
#' fit_seqcox <- coxiv_seq(
#'   formula = survival::Surv(start.new, stop.new, event) ~
#'     reint.base + diameter.base + endoleak.base,
#'   trtformula = reint.base ~ iv.base + diameter.base + endoleak.base,
#'   data = vasc_seqem,
#'   id = "id",
#'   tvtrt = "reint",
#'   iv = "iv",
#'   se_type = "robust",
#'   by_trial = TRUE
#' )
#'
#' print(fit_seqcox)
#'
#' @import survival
#' @export
coxiv_seq <- function(formula,
                      trtformula,
                      data,
                      id,
                      tvtrt,
                      iv,
                      stratify = FALSE,
                      se_type = c("robust", "jackknife", "bootstrap", "none"),
                      B = 50,
                      seed = NULL,
                      ipacw_den_covs = NULL,
                      ipacw_num_covs = NULL,
                      by_trial = TRUE,
                      ...) {

  cl <- match.call()
  se_type <- match.arg(se_type)
  trial_id <- "trial"

  if (!inherits(data, "seqem")) {
    stop("`data` must be a `seqem` object produced by `seqem()`.")
  }

  dat_orig <- data$data_seqorig
  if (!is.data.frame(dat_orig)) {
    stop("`data` must contain `data_seqorig` from `seqem()`.")
  }

  needed_core <- c(id, trial_id, tvtrt, "rownum")
  missing_core <- setdiff(needed_core, names(dat_orig))
  if (length(missing_core) > 0L) {
    stop("Sequential trials data are missing required columns: ",
         paste(missing_core, collapse = ", "), ".")
  }

  if (!inherits(formula, "formula")) {
    stop("`formula` must be a formula.")
  }
  if (!inherits(trtformula, "formula")) {
    stop("`trtformula` must be a formula.")
  }

  trt_base_var <- paste0(tvtrt, ".base")
  iv_base_var  <- paste0(iv, ".base")

  tt_out <- stats::terms(formula)
  rhs_terms_out <- attr(tt_out, "term.labels")
  rhs_vars_out  <- all.vars(formula[[3L]])

  if (!trt_base_var %in% rhs_terms_out && !trt_base_var %in% rhs_vars_out) {
    stop("The RHS of `formula` must include `", trt_base_var, "`.")
  }
  if (tvtrt %in% rhs_vars_out) {
    stop("`formula` uses `", tvtrt, "` without the required `.base` suffix. ",
         "Use `", trt_base_var, "` in the second-stage model.")
  }

  trt_lhs_var <- all.vars(trtformula[[2L]])
  if (length(trt_lhs_var) != 1L || !identical(trt_lhs_var, trt_base_var)) {
    stop("The left-hand side of `trtformula` must be exactly `", trt_base_var, "`.")
  }

  rhs_vars_trt <- all.vars(trtformula[[3L]])
  if (!iv_base_var %in% rhs_vars_trt) {
    stop("The RHS of `trtformula` must include `", iv_base_var, "`.")
  }
  if (iv %in% rhs_vars_trt) {
    stop("`trtformula` uses `", iv, "` without the required `.base` suffix. ",
         "Use `", iv_base_var, "` in the first-stage model.")
  }
  if (tvtrt %in% rhs_vars_trt) {
    warning("`trtformula` contains `", tvtrt, "` without `.base`. ",
            "First-stage predictors should generally use trial-baseline variables.")
  }

  remind_nonbase <- function(vars, context) {
    bad <- vars[vars %in% names(dat_orig) &
                  !grepl("\\.base$", vars) &
                  !vars %in% c(tvtrt, iv, id, trial_id, "R", "R.base")]
    bad <- unique(bad)
    if (length(bad) > 0L) {
      warning(
        context, " contains variables without a `.base` suffix: ",
        paste(bad, collapse = ", "),
        ". In sequential 2SRI, users should usually supply the trial-baseline ",
        "versions of covariates."
      )
    }
  }

  remind_nonbase(rhs_vars_out, "`formula`")
  remind_nonbase(rhs_vars_trt, "`trtformula`")

  base_terms_out <- rhs_terms_out[grepl("\\.base$", rhs_terms_out)]
  strip_base <- function(x) sub("\\.base$", "", x)

  ## ---- first-stage treatment models and control function -----------------
  if ("R.base" %in% names(dat_orig)) {
    warning("Overwriting existing `R.base` column in `data_seqorig`.")
  }
  dat_orig$R.base <- NA_real_

  trial_vec_orig <- dat_orig[[trial_id]]
  trial_vals <- sort(unique(trial_vec_orig))
  n_trials <- length(trial_vals)

  trtfit_by_trial <- vector("list", n_trials)
  names(trtfit_by_trial) <- paste0("trtfit_et", seq_len(n_trials))

  for (k in seq_len(n_trials)) {
    tr_val <- trial_vals[k]
    idx_tr <- trial_vec_orig == tr_val
    dat_k  <- dat_orig[idx_tr, , drop = FALSE]

    if (nrow(dat_k) == 0L) {
      trtfit_by_trial[[k]] <- NULL
      next
    }

    base_rows <- !duplicated(dat_k[[id]])
    dat_k_base <- dat_k[base_rows, , drop = FALSE]

    trt_fit_k <- try(
      stats::glm(trtformula, data = dat_k_base, family = stats::binomial()),
      silent = TRUE
    )

    if (inherits(trt_fit_k, "try-error")) {
      trtfit_by_trial[[k]] <- NULL
      next
    }

    trtfit_by_trial[[k]] <- trt_fit_k

    A_base_k <- dat_k_base[[trt_base_var]]
    p_hat_k  <- stats::fitted(trt_fit_k)
    R_k      <- A_base_k - p_hat_k

    ids_base_k <- dat_k_base[[id]]
    ids_all_k  <- dat_k[[id]]
    map_k      <- match(ids_all_k, ids_base_k)

    if (anyNA(map_k)) {
      stop("Failed to map first-stage residuals back to the stacked data for trial ",
           tr_val, ".")
    }

    dat_orig$R.base[idx_tr] <- R_k[map_k]
  }

  if (anyNA(dat_orig$R.base)) {
    warning("Some rows have missing `R.base` after first-stage estimation.")
  }

  dat_orig$R <- dat_orig$R.base

  ## ---- IPACW for artificial censoring only -------------------------------
  if (is.null(ipacw_den_covs)) {
    den_base_terms <- unique(c(base_terms_out, rhs_vars_trt[grepl("\\.base$", rhs_vars_trt)]))
    den_base_terms <- setdiff(den_base_terms, c(trt_base_var, iv_base_var, "R.base"))
    den_underlying <- strip_base(den_base_terms)
    den_underlying <- den_underlying[den_underlying %in% names(dat_orig)]
    tv_covs <- unique(c(den_underlying, iv, "R"))
    tv_covs <- tv_covs[tv_covs %in% names(dat_orig)]
  } else {
    tv_covs <- unique(c(ipacw_den_covs, iv, "R"))
    tv_covs <- tv_covs[tv_covs %in% names(dat_orig)]
  }

  baseline_covs_num <- NULL
  if (!is.null(ipacw_num_covs)) {
    baseline_covs_num <- ipacw_num_covs[ipacw_num_covs %in% names(dat_orig)]
  }

  dat <- .compute_ipacw_seq_iv(
    dat_seq_orig = dat_orig,
    id_var = id,
    trial_var = trial_id,
    trt_var = tvtrt,
    trt_base_var = trt_base_var,
    rownum_var = "rownum",
    tv_cov_lead = tv_covs,
    baseline_covs_base = baseline_covs_num
  )

  wt_col <- ".coxivseq_wt"
  dat[[wt_col]] <- as.numeric(dat$ipacw.stab.cum)

  ## ---- build Cox formulas ------------------------------------------------
  add_Rbase <- function(formula) {
    f_chr <- paste(deparse(formula), collapse = " ")
    if (grepl("R.base", f_chr, fixed = TRUE)) return(formula)
    parts <- strsplit(f_chr, "~", fixed = TRUE)[[1L]]
    lhs <- trimws(parts[1L])
    rhs <- trimws(parts[2L])
    rhs_new <- if (rhs == "" || rhs == "1") "R.base" else paste(rhs, "R.base", sep = " + ")
    stats::as.formula(paste(lhs, "~", rhs_new), env = environment(formula))
  }

  add_strata <- function(formula, trial_id) {
    f_chr <- paste(deparse(formula), collapse = " ")
    if (grepl("strata(", f_chr, fixed = TRUE)) return(formula)
    parts <- strsplit(f_chr, "~", fixed = TRUE)[[1L]]
    lhs <- trimws(parts[1L])
    rhs <- trimws(parts[2L])
    rhs_new <- if (rhs == "" || rhs == "1") {
      sprintf("strata(%s)", trial_id)
    } else {
      paste(rhs, sprintf("strata(%s)", trial_id), sep = " + ")
    }
    stats::as.formula(paste(lhs, "~", rhs_new), env = environment(formula))
  }

  add_cluster <- function(formula, id_var) {
    f_chr <- paste(deparse(formula), collapse = " ")
    if (grepl("cluster(", f_chr, fixed = TRUE)) return(formula)
    parts <- strsplit(f_chr, "~", fixed = TRUE)[[1L]]
    lhs <- trimws(parts[1L])
    rhs <- trimws(parts[2L])
    rhs_new <- if (rhs == "" || rhs == "1") {
      sprintf("cluster(%s)", id_var)
    } else {
      paste(rhs, sprintf("cluster(%s)", id_var), sep = " + ")
    }
    stats::as.formula(paste(lhs, "~", rhs_new), env = environment(formula))
  }

  form_fit <- add_Rbase(formula)
  if (isTRUE(stratify)) {
    form_fit <- add_strata(form_fit, trial_id)
  }
  form_trial <- add_Rbase(formula)

  ## ---- composite fit -----------------------------------------------------
  fit_main <- survival::coxph(
    formula = form_fit,
    data = dat,
    weights = .coxivseq_wt,
    ...
  )

  coef_hat <- stats::coef(fit_main)
  if (is.null(coef_hat) || length(coef_hat) == 0L) {
    stop("No coefficients were estimated by the composite Cox model.")
  }

  var_hat <- fit_main$var
  fit_used <- fit_main
  jackknife_est <- NULL
  bootstrap_est <- NULL

  if (se_type == "robust") {
    form_rob <- add_cluster(form_fit, id)
    fit_rob <- survival::coxph(
      formula = form_rob,
      data = dat,
      weights = .coxivseq_wt,
      robust = TRUE,
      ...
    )
    coef_hat <- stats::coef(fit_rob)
    var_hat <- fit_rob$var
    fit_used <- fit_rob
  } else if (se_type == "jackknife") {
    id_vec <- dat[[id]]
    ids_unique <- unique(id_vec)
    p <- length(coef_hat)
    jackknife_est <- matrix(NA_real_, nrow = length(ids_unique), ncol = p)
    colnames(jackknife_est) <- names(coef_hat)
    rows_by_id <- split(seq_len(nrow(dat)), id_vec)

    for (j in seq_along(ids_unique)) {
      keep_rows <- unlist(rows_by_id[setdiff(names(rows_by_id), ids_unique[j])],
                          use.names = FALSE)
      dat_j <- dat[keep_rows, , drop = FALSE]
      fit_j <- try(
        survival::coxph(form_fit, data = dat_j, weights = .coxivseq_wt, ...),
        silent = TRUE
      )
      if (!inherits(fit_j, "try-error")) {
        jackknife_est[j, ] <- stats::coef(fit_j)
      }
    }

    valid_rows <- stats::complete.cases(jackknife_est)
    if (sum(valid_rows) > 1L) {
      jk <- jackknife_est[valid_rows, , drop = FALSE]
      m <- nrow(jk)
      jk_centered <- sweep(jk, 2L, coef_hat, FUN = "-")
      var_hat <- (m - 1) / m * crossprod(jk_centered)
    } else {
      warning("Too few valid jackknife refits; reverting to model-based variance.")
    }
  } else if (se_type == "bootstrap") {
    if (!is.null(seed)) set.seed(seed)

    id_vec <- dat[[id]]
    ids_unique <- unique(id_vec)
    n_id <- length(ids_unique)
    p <- length(coef_hat)
    bootstrap_est <- matrix(NA_real_, nrow = B, ncol = p)
    colnames(bootstrap_est) <- names(coef_hat)
    rows_by_id <- split(seq_len(nrow(dat)), id_vec)

    for (b in seq_len(B)) {
      ids_b <- sample(ids_unique, size = n_id, replace = TRUE)
      idx_b <- unlist(rows_by_id[match(ids_b, ids_unique)], use.names = FALSE)
      dat_b <- dat[idx_b, , drop = FALSE]
      fit_b <- try(
        survival::coxph(form_fit, data = dat_b, weights = .coxivseq_wt, ...),
        silent = TRUE
      )
      if (!inherits(fit_b, "try-error")) {
        bootstrap_est[b, ] <- stats::coef(fit_b)
      }
    }

    valid_rows <- stats::complete.cases(bootstrap_est)
    if (sum(valid_rows) > 1L) {
      var_hat <- stats::cov(bootstrap_est[valid_rows, , drop = FALSE])
    } else {
      warning("Too few valid bootstrap refits; reverting to model-based variance.")
    }
  }

  ## ---- trial-specific second-stage fits ---------------------------------
  fit_by_trial <- NULL
  if (isTRUE(by_trial)) {
    trial_vec <- dat[[trial_id]]
    trial_vals2 <- sort(unique(trial_vec))
    fit_by_trial <- vector("list", length(trial_vals2))
    names(fit_by_trial) <- paste0("fit_et", seq_along(trial_vals2))

    for (k in seq_along(trial_vals2)) {
      tr_val <- trial_vals2[k]
      dat_k <- dat[trial_vec == tr_val, , drop = FALSE]
      if (nrow(dat_k) < 1L) {
        fit_by_trial[[k]] <- NULL
        next
      }
      fit_k <- try(
        survival::coxph(form_trial, data = dat_k, weights = .coxivseq_wt, ...),
        silent = TRUE
      )
      if (inherits(fit_k, "try-error")) {
        fit_by_trial[[k]] <- NULL
      } else {
        fit_by_trial[[k]] <- fit_k
      }
    }
  }

  ## ---- summaries --------------------------------------------------------
  se_hat <- sqrt(diag(var_hat))
  z_val  <- coef_hat / se_hat
  p_val  <- 2 * stats::pnorm(abs(z_val), lower.tail = FALSE)

  conf_int <- data.frame(
    `2.5%` = coef_hat - stats::qnorm(0.975) * se_hat,
    `97.5%` = coef_hat + stats::qnorm(0.975) * se_hat,
    check.names = FALSE
  )
  rownames(conf_int) <- names(coef_hat)

  coef_df <- data.frame(
    logHR = coef_hat,
    HR = exp(coef_hat),
    se = se_hat,
    Z = z_val,
    p.val = p_val,
    check.names = FALSE
  )
  rownames(coef_df) <- names(coef_hat)

  n_ids <- length(unique(dat[[id]]))
  n_trials2 <- length(unique(dat[[trial_id]]))

  res <- list(
    call = cl,
    formula = form_fit,
    trtformula = trtformula,
    tvtrt = tvtrt,
    iv = iv,
    fit = fit_used,
    coef = coef_df,
    conf_int = conf_int,
    vcov_mat = var_hat,
    n = nrow(dat),
    n_ids = n_ids,
    n_trials = n_trials2,
    n_event = if (!is.null(fit_used$nevent)) fit_used$nevent else sum(dat[[all.vars(formula[[2L]])[3L]]], na.rm = TRUE),
    stratify = stratify,
    se_type = se_type,
    B = if (se_type == "bootstrap") B else NULL,
    jackknife = jackknife_est,
    bootstrap = bootstrap_est,
    data_ste = dat,
    trtfit_by_trial = trtfit_by_trial,
    by_trial = by_trial,
    fit_by_trial = fit_by_trial
  )

  class(res) <- "coxivseq"
  res
}


.compute_ipacw_seq_iv <- function(dat_seq_orig,
                                  id_var,
                                  trial_var,
                                  trt_var,
                                  trt_base_var,
                                  rownum_var = "rownum",
                                  tv_cov_lead = NULL,
                                  baseline_covs_base = NULL) {

  stopifnot(id_var %in% names(dat_seq_orig),
            trial_var %in% names(dat_seq_orig),
            trt_var %in% names(dat_seq_orig),
            trt_base_var %in% names(dat_seq_orig),
            rownum_var %in% names(dat_seq_orig))

  ord <- order(dat_seq_orig[[id_var]], dat_seq_orig[[trial_var]], dat_seq_orig[[rownum_var]])
  dat_seq_orig <- dat_seq_orig[ord, , drop = FALSE]

  g <- interaction(dat_seq_orig[[id_var]], dat_seq_orig[[trial_var]], drop = TRUE)

  dat_seq_orig$A.lead1 <- ave(dat_seq_orig[[trt_var]], g,
                              FUN = function(x) c(x[-1L], NA))

  dat_seq_orig$Anext.equal.to.baseline <- ifelse(
    is.na(dat_seq_orig$A.lead1),
    NA_integer_,
    ifelse(dat_seq_orig$A.lead1 == dat_seq_orig[[trt_base_var]], 1L, 0L)
  )

  lead_cov_names <- character(0L)
  if (!is.null(tv_cov_lead) && length(tv_cov_lead) > 0L) {
    for (v in tv_cov_lead) {
      if (!v %in% names(dat_seq_orig)) next
      newname <- paste0(v, ".lead1")
      dat_seq_orig[[newname]] <- ave(dat_seq_orig[[v]], g,
                                     FUN = function(x) c(x[-1L], NA))
      lead_cov_names <- c(lead_cov_names, newname)
    }
  }

  keep_idx <- dat_seq_orig[[trt_var]] == dat_seq_orig[[trt_base_var]]
  dat_seq <- dat_seq_orig[keep_idx, , drop = FALSE]

  dat_seq$wt.denom <- 1
  dat_seq$wt.num <- 1

  mask_est <- (dat_seq[[trt_base_var]] == 0) & !is.na(dat_seq$Anext.equal.to.baseline)

  if (any(mask_est)) {
    est_data <- dat_seq[mask_est, , drop = FALSE]

    form_denom <- if (length(lead_cov_names) > 0L) {
      stats::as.formula(paste("Anext.equal.to.baseline ~", paste(lead_cov_names, collapse = " + ")))
    } else {
      Anext.equal.to.baseline ~ 1
    }

    wt.mod.denom <- stats::glm(form_denom, family = stats::binomial(), data = est_data)
    dat_seq$wt.denom[mask_est] <- stats::predict(wt.mod.denom, newdata = est_data, type = "response")

    if (is.null(baseline_covs_base) || length(baseline_covs_base) == 0L) {
      form_num <- Anext.equal.to.baseline ~ 1
    } else {
      base_covs_in_data <- baseline_covs_base[baseline_covs_base %in% names(dat_seq)]
      if (length(base_covs_in_data) == 0L) {
        form_num <- Anext.equal.to.baseline ~ 1
      } else {
        rhs_num <- paste0(
          "as.factor(", rownum_var, ")*(",
          paste(base_covs_in_data, collapse = " + "),
          ")"
        )
        form_num <- stats::as.formula(paste("Anext.equal.to.baseline ~", rhs_num))
      }
    }

    wt.mod.num <- stats::glm(form_num, family = stats::binomial(), data = est_data)
    dat_seq$wt.num[mask_est] <- stats::predict(wt.mod.num, newdata = est_data, type = "response")
  }

  dat_seq$wt.denom <- ifelse(dat_seq[[trt_base_var]] == 1, 1, dat_seq$wt.denom)
  dat_seq$wt.num   <- ifelse(dat_seq[[trt_base_var]] == 1, 1, dat_seq$wt.num)

  dat_seq$ipacw.stab <- dat_seq$wt.num / dat_seq$wt.denom

  g2 <- interaction(dat_seq[[id_var]], dat_seq[[trial_var]], drop = TRUE)
  dat_seq$ipacw.stab.lag <- ave(dat_seq$ipacw.stab, g2,
                                FUN = function(x) c(1, x[-length(x)]))
  dat_seq$ipacw.stab.cum <- ave(dat_seq$ipacw.stab.lag, g2, FUN = cumprod)

  dat_seq
}

#' vascular: longitudinal vascular surgery follow-up data
#'
#' A synthetic longitudinal dataset designed to mimic repeated follow-up in a
#' vascular surgery study. The data are stored in long start-stop format for
#' survival analyses with a time-varying exposure and a time-varying binary
#' prognostic factor. In particular, `reint` indicates reintervention surgery
#' status over time and `endoleak` indicates time-varying endoleak status over
#' time. Both are coded so that once the status becomes 1 it remains 1 at later
#' visits.
#'
#' @format A data frame with variables:
#' * `id` : Unique patient identifier
#' * `surv.time` : Observed overall follow-up time to death or censoring
#' * `death` : Terminal event indicator
#' * `time.start` : Interval start time
#' * `time.stop` : Interval stop time
#' * `event` : Counting-process event indicator for the interval
#' * `reint` : Time-varying reintervention status at the current visit
#' * `endoleak` : Time-varying endoleak status at the current visit
#' * `iv` : Baseline instrumental variable for reintervention propensity
#' * `smoke` : Baseline smoking status indicator
#' * `age` : Baseline age
#' * `diameter` : Baseline aneurysm diameter
#' * `gender` : Baseline gender indicator
#' * `hyper` : Baseline hypertension indicator
#' * `obese` : Baseline obesity indicator
#' * `reintlag1` : Lagged reintervention status from the previous interval
#' * `visit` : Visit number
#' @source Synthetic data generated to mimic the distributional features of a
#'   vascular surgery longitudinal follow-up study.
#' @usage data(vascular)
"vascular"

#' VitD: cohort data on vitamin D and mortality
#'
#' This dataset is mirrored from the \pkg{ivtools} package (version 2.3.0).
#' The upstream documentation notes the data originate from a real cohort study
#' and were modified for public availability.
#'
#' @format A data frame with variables:
#' * `age` : Age at baseline
#' * `filaggrin` : Indicator of filaggrin mutation
#' * `vitd` : Serum 25-OH-D (nmol/L) at baseline
#' * `time` : Follow-up time
#' * `death` : Death indicator during follow-up
#' @source Mirrored from \pkg{ivtools} (LGPL (>= 3)).
#' @references Sjolander, Arvid and Martinussen, Torben. "Instrumental Variable Estimation with the R Package ivtools" Epidemiologic Methods, vol. 8, no. 1, 2019, pp. 20180024.
#' @usage data(VitD)
"VitD"

#' Print method for seqcox objects
#'
#' @param x An object of class `seqcox` returned by [seqcox()].
#' @param digits Number of decimal places for printed coefficient summaries.
#' @param ... Ignored.
#'
#' @return The input object `x`, invisibly. Called for its side effect of printing a summary to the console.
#'
#' @export
print.seqcox <- function(x, digits = 4, ...) {
  if (!inherits(x, "seqcox")) {
    return(NextMethod())
  }

  title <- "Sequential Cox analysis via sequential trials emulation"
  cat(title, "\n", sep = "")
  cat(strrep("=", nchar(title)), "\n", sep = "")

  cat(sprintf("Composite analysis across %d emulated trials \n", x$n_trials))
  cat(sprintf("Stratified by trial : %s\n", if (isTRUE(x$stratify)) "yes" else "no"))
  cat(sprintf("Number of individuals : %d\n", x$n_ids))
  cat(sprintf("Number of events: %d\n", x$n_event))
  cat(sprintf("SE type: %s\n", x$se_type))

  n_fit_by_trial <- 0L
  if (!is.null(x$fit_by_trial)) {
    n_fit_by_trial <- sum(!vapply(x$fit_by_trial, is.null, logical(1)))
  }
  cat(sprintf("Trial-specific fits: %d model(s) available in [ $fit_by_trial ]\n",
              n_fit_by_trial))

  cat(strrep("~", nchar(title)), "\n", sep = "")
  cat("Coefficients (composite fit):\n")

  tab <- x$coef[, c("logHR", "HR", "se", "p.val"), drop = FALSE]
  tab_chr <- as.matrix(tab)
  tab_chr[,] <- formatC(as.matrix(tab), format = "f", digits = digits)
  print(noquote(tab_chr), quote = FALSE, right = TRUE)

  cat(strrep("-", nchar(title)), "\n", sep = "")
  cat("A `seqcox` object\n")
  invisible(x)
}

#' Print the output from a `coxivtsrif` model fit.
#'
#' Provide succinct and organized output for a `coxivtsrif` model, such as
#' `coxiv_tsrif()`.
#'
#' @param x an object of class `"coxivtsrif"`, usually created by
#'   `coxiv_tsrif()`.
#' @param digits the number of decimal places to display for coefficient
#'   summaries and hazard ratios.
#' @param ... additional arguments (currently unused).
#'
#' @return The input object `x`, invisibly. Called for its side effect of printing a summary to the console.
#'
#' @examples
#' data("VitD", package = "surviv")
#' fit_tsrif <- surviv::coxiv_tsrif(
#'   surv = "time",
#'   cens = "death",
#'   covs = c("age"),
#'   trt = "vitd",
#'   iv = "filaggrin",
#'   tchange = NULL,
#'   data = VitD,
#'   fdist = "gaussian",
#'   bootvar = FALSE,
#'   B = 20,
#'   tvareff = FALSE
#' )
#' print(fit_tsrif)
#'
#' @export
print.coxivtsrif <- function(x, digits = 3L, ...) {
  if (!inherits(x, "coxivtsrif")) {
    stop("Input must be of class 'coxivtsrif'.")
  }

  .print_text <- function(obj, ...) {
    txt <- utils::capture.output(print(obj, ...))
    cat(paste0(txt, collapse = "\n"), "\n")
  }

  .fmt_noquote <- function(mat, digits) {
    mat <- as.matrix(mat)
    out <- matrix(
      formatC(mat, format = "f", digits = digits),
      nrow = nrow(mat),
      ncol = ncol(mat),
      dimnames = dimnames(mat)
    )
    noquote(out)
  }

  .extract_trt_coef <- function(obj) {
    trt_coef <- obj$trt_coef
    if (!is.null(trt_coef)) {
      trt_coef <- as.data.frame(trt_coef, check.names = FALSE)
      nm <- names(trt_coef)
      map <- c(
        "Estimate" = "logHR",
        "Std.Error" = "se",
        "Z.value" = "Z",
        "P(>|z|)" = "pval",
        "logHR" = "logHR",
        "se" = "se",
        "Z" = "Z",
        "pval" = "pval"
      )
      nm2 <- unname(map[nm])
      names(trt_coef) <- ifelse(is.na(nm2), nm, nm2)
      if (!all(c("logHR", "se") %in% names(trt_coef))) {
        stop("`x$trt_coef` must contain `logHR` and `se` columns.")
      }
      trt_coef$logHR <- as.numeric(trt_coef$logHR)
      trt_coef$se <- as.numeric(trt_coef$se)
      if ("Z" %in% names(trt_coef)) trt_coef$Z <- as.numeric(trt_coef$Z)
      if ("pval" %in% names(trt_coef)) trt_coef$pval <- as.numeric(trt_coef$pval)
      return(trt_coef)
    }

    # Backward compatibility with older objects
    est <- obj$est_coef
    if (is.null(est)) {
      stop("Cannot find treatment-effect summary in `x$trt_coef` or `x$est_coef`.")
    }

    ci <- obj$conf_int
    if (is.null(ci)) {
      stop("Cannot find confidence intervals in `x$conf_int`.")
    }

    if (is.matrix(ci) || is.data.frame(ci)) {
      ci_mat <- as.matrix(ci)
    } else {
      ci_mat <- matrix(ci, nrow = length(est), byrow = TRUE)
    }

    est_num <- as.numeric(est)
    if (is.null(names(est)) && !is.null(rownames(ci_mat))) names(est_num) <- rownames(ci_mat)

    se_num <- rep(NA_real_, length(est_num))
    if (!is.null(obj$variance)) {
      se_num <- sqrt(as.numeric(obj$variance))
    }

    z_num <- est_num / se_num
    p_num <- 2 * stats::pnorm(abs(z_num), lower.tail = FALSE)

    trt_coef <- data.frame(
      logHR = est_num,
      se = se_num,
      Z = z_num,
      pval = p_num,
      row.names = if (!is.null(names(est_num))) names(est_num) else NULL,
      check.names = FALSE
    )
    trt_coef
  }

  .extract_confint <- function(obj, rn = NULL) {
    ci <- obj$conf_int
    if (is.null(ci)) {
      stop("Cannot find confidence intervals in `x$conf_int`.")
    }

    if (is.vector(ci) && !is.list(ci)) {
      if (length(ci) %% 2L != 0L) stop("`x$conf_int` has invalid length.")
      ci <- matrix(ci, ncol = 2, byrow = TRUE)
    }

    ci <- as.data.frame(ci, check.names = FALSE)
    if (ncol(ci) < 2L) stop("`x$conf_int` must have two columns.")
    ci <- ci[, 1:2, drop = FALSE]
    names(ci) <- c("2.5%", "97.5%")
    ci[[1]] <- as.numeric(ci[[1]])
    ci[[2]] <- as.numeric(ci[[2]])

    if (!is.null(rn) && is.null(rownames(ci)) && nrow(ci) == length(rn)) {
      rownames(ci) <- rn
    }
    ci
  }

  .compact_cox_table <- function(model, trt_name = NULL, trt_override = NULL) {
    beta <- stats::coef(model)
    if (is.null(beta) || length(beta) == 0L) {
      return(NULL)
    }
    beta <- as.numeric(beta)
    names(beta) <- names(stats::coef(model))

    se <- tryCatch(
      sqrt(diag(stats::vcov(model))),
      error = function(e) sqrt(diag(model$var))
    )
    se <- as.numeric(se)
    names(se) <- names(stats::coef(model))

    tab <- data.frame(
      logHR = beta,
      se = se,
      HR = exp(beta),
      row.names = names(beta),
      check.names = FALSE
    )

    if (!is.null(trt_name) && !is.null(trt_override) && trt_name %in% rownames(tab)) {
      tab[trt_name, "logHR"] <- as.numeric(trt_override[["logHR"]])
      tab[trt_name, "se"] <- as.numeric(trt_override[["se"]])
      tab[trt_name, "HR"] <- exp(as.numeric(trt_override[["logHR"]]))
    }

    tab
  }

  .pick_n_events <- function(obj) {
    n_obs <- NA_integer_
    n_evt <- NA_integer_

    if (!is.null(obj$surv_curve$n)) {
      n_obs <- as.integer(obj$surv_curve$n[1])
    } else if (!is.null(obj$out_model$n)) {
      n_obs <- as.integer(obj$out_model$n[1])
    } else if (!is.null(obj$out_model1$n)) {
      n_obs <- as.integer(obj$out_model1$n[1])
    }

    if (!is.null(obj$surv_curve$n.event)) {
      n_evt <- as.integer(sum(obj$surv_curve$n.event, na.rm = TRUE))
    } else if (!is.null(obj$out_model$nevent)) {
      n_evt <- as.integer(obj$out_model$nevent[1])
    } else if (!is.null(obj$out_model1$nevent) && !is.null(obj$out_model2$nevent)) {
      # For time-varying models, `surv_curve$n.event` is preferred; otherwise use summed events.
      n_evt <- as.integer(obj$out_model1$nevent[1] + obj$out_model2$nevent[1])
    }

    list(n = n_obs, events = n_evt)
  }

  trt_name <- if (!is.null(x$input) && !is.null(x$input$trt)) x$input$trt else NULL
  trt_coef <- .extract_trt_coef(x)
  if (is.null(trt_name) || !nzchar(trt_name)) {
    trt_name <- rownames(trt_coef)[1]
    trt_name <- sub("\\.t[12]$", "", trt_name)
  }
  conf_df <- .extract_confint(x, rn = rownames(trt_coef))
  n_info <- .pick_n_events(x)

  cat("TSRI-Frailty instrumental variable analysis of Cox model\n")
  cat(strrep("=", 56), "\n")

  if (isTRUE(x$input$tvareff)) {
    cat("Call:\n")
    cat("- First-stage (treatment) model:\n")
    cat(paste(deparse(x$input$trtformula), collapse = " "), "\n")
    cat("- Second-stage (outcome) model:\n")
    cat(paste(deparse(x$input$formula), collapse = " "), "\n")
    cat(sprintf("* Time-varying treatment effect updates at t = %s\n", format(x$input$tchange)))
    cat(strrep("-", 56), "\n")

    if (!is.na(n_info$n)) cat(sprintf("Sample size: %d\n", n_info$n))
    if (!is.na(n_info$events)) cat(sprintf("Number of events: %d\n", n_info$events))
    cat(strrep("-", 56), "\n")

    rn <- rownames(trt_coef)
    if (is.null(rn) || nrow(trt_coef) < 2L) {
      rn <- c(paste0(trt_name, ".t1"), paste0(trt_name, ".t2"))
      rownames(trt_coef) <- rn
      if (nrow(conf_df) == 2L && is.null(rownames(conf_df))) rownames(conf_df) <- rn
    }

    cat(sprintf("Conditional treatment effect (%s):\n", trt_name))
    cat(sprintf("# Period 1 # (t<%.*f):\n", digits, as.numeric(x$input$tchange)))
    cat(sprintf(
      "- Estimate: %.*f [%.*f, %.*f]\n",
      digits, trt_coef[1, "logHR"], digits, conf_df[1, 1], digits, conf_df[1, 2]
    ))
    cat(sprintf(
      "- Hazard ratio: %.*f [%.*f, %.*f]\n",
      digits, exp(trt_coef[1, "logHR"]), digits, exp(conf_df[1, 1]), digits, exp(conf_df[1, 2])
    ))
    cat(strrep("~", 22), "\n")
    cat(sprintf("# Period 2 # (t>%.*f):\n", digits, as.numeric(x$input$tchange)))
    cat(sprintf(
      "- Estimate: %.*f [%.*f, %.*f]\n",
      digits, trt_coef[2, "logHR"], digits, conf_df[2, 1], digits, conf_df[2, 2]
    ))
    cat(sprintf(
      "- Hazard ratio: %.*f [%.*f, %.*f]\n",
      digits, exp(trt_coef[2, "logHR"]), digits, exp(conf_df[2, 1]), digits, exp(conf_df[2, 2])
    ))
    cat(strrep("-", 56), "\n")

    cat("Coefficients:\n")
    cat("# Period 1 #:\n")
    tab1 <- .compact_cox_table(x$out_model1, trt_name = trt_name, trt_override = trt_coef[1, ])
    if (!is.null(tab1)) .print_text(.fmt_noquote(tab1, digits = digits), right = TRUE)

    cat(strrep("~", 13), "\n")
    cat("# Period 2 #:\n")
    tab2 <- .compact_cox_table(x$out_model2, trt_name = trt_name, trt_override = trt_coef[2, ])
    if (!is.null(tab2)) .print_text(.fmt_noquote(tab2, digits = digits), right = TRUE)
    cat(strrep("-", 56), "\n")
  } else {
    cat("Call:\n")
    cat("- First-stage (treatment) model:\n")
    cat(paste(deparse(x$input$trtformula), collapse = " "), "\n")
    cat("- Second-stage (outcome) model:\n")
    cat(paste(deparse(x$input$formula), collapse = " "), "\n")
    cat("* Time-constant treatment effect\n")
    cat(strrep("-", 56), "\n")

    if (!is.na(n_info$n)) cat(sprintf("Sample size: %d\n", n_info$n))
    if (!is.na(n_info$events)) cat(sprintf("Number of events: %d\n", n_info$events))
    cat(strrep("-", 56), "\n")

    cat(sprintf("Conditional treatment effect (%s):\n", trt_name))
    cat(sprintf(
      "- Estimate: %.*f [%.*f, %.*f]\n",
      digits, trt_coef[1, "logHR"], digits, conf_df[1, 1], digits, conf_df[1, 2]
    ))
    cat(sprintf(
      "- Hazard ratio: %.*f [%.*f, %.*f]\n",
      digits, exp(trt_coef[1, "logHR"]), digits, exp(conf_df[1, 1]), digits, exp(conf_df[1, 2])
    ))
    cat(strrep("-", 56), "\n")

    cat("Coefficients:\n")
    tab <- .compact_cox_table(x$out_model, trt_name = trt_name, trt_override = trt_coef[1, ])
    if (!is.null(tab)) .print_text(.fmt_noquote(tab, digits = digits), right = TRUE)
    cat(strrep("-", 56), "\n")
  }

  if (!is.null(x$iv_diag$f_stat)) {
    cat("Nested exclusion test for weak IV:\n")
    cat(sprintf(
      "- F-statistics: %.2f%s\n",
      as.numeric(x$iv_diag$f_stat),
      ifelse(as.numeric(x$iv_diag$f_stat) > 10, " [> 10.0]", " [< 10.0]")
    ))
    cat(strrep("-", 56), "\n")
  }

  cat("A `coxivtsrif` object\n")
  invisible(x)
}

#' Two-stage residual inclusion-frailty (TSRI-F) instrumental variable analysis of Cox model
#'
#' Flexible instrumental analysis of Cox model in a novel **T**wo-**S**tage **R**esidual **I**nclusion with **F**railty (TSRI-F) framework when treatment \eqn{A} and mortality \eqn{D} are subject to unmeasured confounding.
#' Allow estimation of time constant treatment effect (Martinez-Camblor et al. (2019)) and time-varying treatment effect (Martinez-Camblor et al. (2019))
#' in terms of log hazard ratio (log-HR). A set of instrumental variables \eqn{Z} and covariates for adjustment \eqn{X} (i.e., measured confounders) are required to estimate treatment effects \eqn{\beta^{(1)}_a} and \eqn{\beta^{(2)}_a} before and after the pre-specified time `tchange`.
#'
#' This function performs two-stage residual inclusion IV analysis of the Cox model with individual frailty
#' when the treatment effect \eqn{\beta_a} is constant over time (`tvareff = FALSE`) or \eqn{\beta_a (t)} changes value at some time _t_ (`tvareff = TRUE`).
#' The IV assumptions are that 1) relevance \eqn{Z \not\amalg A | U,X}, 2) exclusion restriction \eqn{Z\amalg (T,D)|A, U,X}, and 3) randomization \eqn{Z\amalg U|X} with \eqn{U} the unmeasured confounder.
#' TSRI-F posits a first-stage linear model that generates the treatment: \deqn{A = \gamma_0+ \gamma_1Z+\gamma_2U+\gamma'X+\epsilon}
#' and a second-stage outcome model follows Cox proportional hazards form: \deqn{\lambda(t|X,U)=\lambda_0(t)\exp\{\beta_aA+\beta_uU+\beta'X\}}
#' where \eqn{\lambda_0} is the baseline hazard function. The first stage of TSRIF procedure for constant treatment effect estimates _control function_ \eqn{\widehat{R}} from residual of treatment model:
#' \eqn{\widehat{R} = A - \hat{\gamma}_0 -\hat{\gamma}_1Z-\hat{\gamma}'X}, and fit a second-stage Cox model with user-specified individual frailty \eqn{\phi} while adjusting for \eqn{\widehat{R}}.
#'
#' For more intricate case of time-varying effect, the first stage proceeds identically to compute \eqn{\widehat{R}^{(1)} = A - \hat{\gamma}_0 -\hat{\gamma}_1Z-\hat{\gamma}'X}.
#' Second stage involves fitting two Cox models to estimate \eqn{\beta_a^{(1)} = \beta_a(t), t<} `tchange` and \eqn{\beta_a^{(2)} = \beta_a(t), t\ge} `tchange`.
#' The first-period Cox model is fitted with frailty term \eqn{\hat{\phi_1}} and including \eqn{\widehat{R}^{(1)}} as a covariate.
#' Then, update the control function \eqn{\widehat{R}^{(2)} = \widehat{R}^{(1)} + \hat{\beta}^{(1), -1}_r \hat{\phi_1}} for the second period, which are then included as a covariate in the second-period Cox model
#' with a new frailty term.
#'
#' This approach estimates conditional treatment effect of `trt`. Coefficient estimates of covariates other than `trt` are also estimated and readily available, but these estimates are
#' suggestive of their associational relations with survival outcome, and causal conclusions regarding these covariates with outcome should be cautious.
#'
#' @usage coxiv_tsrif(surv, cens, trt, iv, covs, data, tchange = NULL,
#'             tvareff = FALSE, fdist = "gamma", bootvar = FALSE, B = 50)
#'
#' @param surv a string indicating variable name of survival time \eqn{T}.
#' @param cens a string indicating variable name of event indicator \eqn{D}: 1 for terminal event occurrence and 0 for right censor.
#' @param trt a string indicating the treatment variable name \eqn{A}. Binary or continuous treatment supported.
#' @param iv a string or vector of strings indicating instrumental variable(s) \eqn{Z}. Binary or continuous IV supported.
#' @param covs a string or vector of strings indicating set of measured covariates \eqn{X}.
#' @param data a \link[base]{data.frame} containing all variables needed.
#' @param tchange a positive numeric value specifying the time point at which the treatment effect changes value.
#' @param fdist the frailty distribution, default is `gamma`. Other options include `gaussian` or `t` distribution. Read more about frailty distribution specification at \link[survival]{frailty}.
#' @param bootvar logical; if `TRUE`, bootstrap variance estimation is performed. Otherwise, analytical variance is provided from frailty Cox model.
#' @param B number of bootstrap iterations if `bootvar` is `TRUE`.
#' @param tvareff logical; by default `FALSE` and estimate a time-constant treatment effect model. If `TRUE`, a time-varying treatment effect model is applied.
#'
#' @return A `coxivtsrif` object with following components:
#' * `input` : collected input and derived model formulas used for the TSRI-F fit.
#' * `trt_coef` : a `data.frame` of treatment-effect estimates with columns `logHR`, `se`, `Z`, and `p.val`. If `tvareff = TRUE`, it contains one row for each treatment-effect period.
#' * `conf_int` : a `data.frame` for 95% confidence interval(s) for estimated treatment effect (s) from the choice of standard error estimator.
#' * `ctrl_func` : the estimated control function \eqn{\widehat{R}}, or `data.frame` of \eqn{\widehat{R}^1} and \eqn{\widehat{R}^2} if `tvareff = TRUE`.
#' * `trt_model` : a \link[stats]{lm} object fit of the first-stage treatment model.
#' * `out_model` : a \link[survival]{coxph} object fit of the second-stage outcome Cox model. If `tvareff = TRUE`, Cox models for first period `out_model1` and second period `out_model2` are both returned.
#' * `est_boot` :  a vector of length `B` of bootstrap estimates from each iteration. If `bootvar = FALSE`, returns `NULL`.
#' * `data_long` :  a `data.frame` of data for model fit transformed into start-stop form for two-period Cox estimation, if `tvareff = TRUE`.
#' * `surv_curve`: a \link[survival]{survfit} object for fitted survival curve using complete data.
#' * `iv_diag` : a list of IV diagnostics for weak-IV assessment, including the nested ANOVA comparison and F-statistic.
#'
#' @examples
#' data("VitD", package = "surviv")
#'
#' fit_tsrif1 <- coxiv_tsrif(surv = "time", cens = "death", covs = c("age"),
#'                           trt = "vitd", iv = "filaggrin", tchange = NULL,
#'                           data = VitD, fdist = "gaussian", bootvar = FALSE,
#'                           B = 20, tvareff = FALSE)
#' fit_tsrif1$trt_coef
#'
#' fit_tsrif2 <- coxiv_tsrif(surv = "time", cens = "death", covs = c("age"),
#'                           trt = "vitd", iv = "filaggrin", tchange = 5,
#'                           data = VitD, fdist = "gaussian", bootvar = FALSE,
#'                           B = 20, tvareff = TRUE)
#' fit_tsrif2$trt_coef
#'
#' @import survival stats
#'
#' @references
#' 1. Pablo Martinez-Camblor, Todd Mackenzie, Douglas O Staiger, Philip P Goodney, A James O'Malley, Adjusting for bias introduced by instrumental variable estimation in the Cox proportional hazards model, *Biostatistics*, Volume 20, Issue 1, January 2019, Pages 80-96,
#'
#' 2. Pablo Martinez-Camblor, Todd A. MacKenzie, Douglas O. Staiger, Phillip P. Goodney, A. James O'Malley, An Instrumental Variable Procedure for Estimating Cox Models with Non-Proportional Hazards in the Presence Of Unmeasured Confounding, *Journal of the Royal Statistical Society Series C: Applied Statistics*, Volume 68, Issue 4, August 2019, Pages 985-1005
#' @export
#'
#'

coxiv_tsrif <- function(surv, cens, trt, iv, covs, data, tchange = NULL, tvareff = FALSE,
                        fdist = "gamma", bootvar = FALSE, B = 50) {
  # Input validation
  if (missing(surv) || missing(cens) || missing(covs) || missing(trt) || missing(iv) || missing(data)) {
    stop("All arguments 'surv', 'covs', 'trt', 'iv', and 'data' must be provided.")
  }
  if (!is.character(surv) || length(surv) != 1) stop("'surv' must be a string for survival time variable")
  if (!is.character(cens) || length(cens) != 1) stop("'cens' must be a string for censoring/mortality variable.")
  if (!is.data.frame(data)) stop("'data' must be a data frame.")
  if (!is.character(trt) || length(trt) != 1) stop("'trt' must be a string for treatment variable.")
  if (!is.null(covs) && !is.character(covs)) stop("'covs' must be a string or a vector of strings for confounder(s).")
  if (!(is.character(iv) && (length(iv) >= 1))) {
    stop("'iv' must be a string or a vector of strings for instrumental variable(s).")
  }
  if (!(surv %in% colnames(data))) stop(paste("Survival time", surv, "not found in the data."))
  if (!(cens %in% colnames(data))) stop(paste("Censoring indicator", cens, "not found in the data."))
  if (!(trt %in% colnames(data))) stop(paste("Treatment variable", trt, "not found in the data."))
  if (!all(iv %in% colnames(data))) stop(paste("Instrumental variable(s)", iv, "not found in the data."))
  if (!all(covs %in% colnames(data))) stop("Some confounders are not found in the data.")
  if (tvareff & (is.null(tchange) || tchange < 0 || !is.numeric(tchange))) stop("A valid change time 'tchange' is needed for time-varying treatment effect estimation.")
  if (!tvareff & !is.null(tchange)) message("~ The supplied 'tchange' value ignored since 'tvareff' is FALSE.")
  if (!bootvar & !is.null(B)) message("~ Bootstrap iterations 'B' ignored since 'bootvar' was set to FALSE.")
  if (bootvar && (is.null(B) || !is.numeric(B) || B <= 0)) stop("If bootvar = TRUE, you must specify a positive numeric value for B.")

  # Construct two-stage formulae
  formula <- as.formula(paste("Surv(", surv, ",", cens, ") ~",
                              paste(c(trt, if (!is.null(covs)) covs else NULL), collapse = " + ")))

  trtformula <- as.formula(paste(trt, "~",
                                 paste(c(iv, if (!is.null(covs)) covs else NULL), collapse = " + ")))

  input <- as.list(environment())

  # Handle missing data
  complete_data <- na.omit(data)
  if (nrow(complete_data) < nrow(data)) {
    message("~ Missing data found. Complete case analysis is performed.")
  }

  # Delegate to the appropriate function based on tvareff
  result <- if (tvareff) {
    message("~ A TSRI-F Cox model with *time-varying* treatment effect is estimated.")
    est_tsrif_tvar(formula = formula, trtformula = trtformula, data = complete_data,
                   trt = trt, iv = iv, t = tchange, fdist = fdist, bootvar = bootvar, B = B)
  } else {
    message("~ A TSRI-F Cox model with *constant* treatment effect is estimated.")
    est_tsrif_tfix(formula = formula, trtformula = trtformula, data = complete_data,
                   trt = trt, iv = iv, fdist = fdist, bootvar = bootvar, B = B)
  }

  # Instrumental variable relevance diagnostics
  iv_diag <- iv_strgth(covs = covs, trt = trt, iv = iv, data = complete_data)
  if (iv_diag$f_stat < 10) message("Attention! Possible weak IV: nested anova has F < 10.0 for IV-exclusion treatment model.")

  # Append additional output elements
  result$input <- input
  result$iv_diag <- iv_diag

  # Organize results
  class(result) <- c("coxivtsrif")

  return(result)
}

# TSRIF Cox model for time-constant effect

#' @noRd
est_tsrif_tfix <- function(formula, trtformula, data, trt, iv, fdist = "gamma",
                           bootvar = FALSE, B = NULL) {
  # Ensure the necessary libraries are loaded
  requireNamespace("survival")
  requireNamespace("stats")

  # Extract survival data
  data_outcome_model <- model.frame(formula, data = data)
  if (!inherits(data_outcome_model[,1], "Surv")) stop("The left-hand side of the outcome formula must be a 'Surv' object.")
  survtimes <- data_outcome_model[, 1][, 1]  # Extract survival times
  eventind <- data_outcome_model[, 1][, 2]  # Extract event indicators

  # Estimate survival function
  surv_curve <- survfit(Surv(survtimes, eventind) ~ 1)

  # Extract variables from trtformula
  trt_model <- lm(trtformula, data = data)
  Resid <- residuals(trt_model)

  # Add residuals to data for the second stage
  data$Resid <- Resid

  # Fit the second-stage Cox model with frailty
  cox_model <- coxph(
    formula = as.formula(paste0(
      deparse(formula),
      " + Resid + frailty(as.factor(1:",nrow(data),"), dist = '", fdist, "')"
    )),
    data = data
  )

  # Extract coefficients from outcome model
  est <- coef(cox_model)[trt]
  #names(est)[length(est)] <- "CtrlFunc"

  # Calculate variance
  if (!bootvar) {
    # Use variance from the coxph model
    variance <- vcov(cox_model)[trt, trt]
  } else {
    # Bootstrap variance estimation
    N <- nrow(data)
    boot_coefs <- numeric(B)
    message("... Bootstrap simulation starts ...")

    for (b in 1:B) {
      # Bootstrap resampling
      boot_indices <- sample(1:N, replace = TRUE)
      boot_data <- data[boot_indices, ]

      # First-stage model on bootstrap sample
      boot_trt_model <- lm(trtformula, data = boot_data)
      boot_data$Resid <- residuals(boot_trt_model)

      # Second-stage model on bootstrap sample
      boot_cox_model <- coxph(
        formula = as.formula(paste0(
          deparse(formula),
          " + Resid + frailty(as.factor(1:",nrow(data),"), dist = '", fdist, "')"
        )),
        data = boot_data
      )

      # Store treatment effect estimate
      boot_coefs[b] <- coef(boot_cox_model)[trt]
    }
    message("... Bootstrap simulation ends ...")

    # Variance of bootstrap estimates
    variance <- var(boot_coefs)
  }

  # 95% Confidence Interval for treatment effect
  upper <- est + qnorm(0.975) * sqrt(variance)
  lower <- est - qnorm(0.975) * sqrt(variance)
  conf_int <- c(lower, upper)
  names(conf_int) <- c("2.5%", "97.5%")

  trt_coef <- data.frame(
    logHR = unname(est),
    se = sqrt(variance),
    Z = unname(est) / sqrt(variance),
    p.val = 2 * pnorm(abs(unname(est) / sqrt(variance)), lower.tail = FALSE),
    row.names = trt,
    check.names = FALSE
  )

  # Return results
  return(list(
    trt_coef = trt_coef,
    conf_int = conf_int,
    trt_model = trt_model,
    ctrl_func = Resid,
    out_model = cox_model,
    est_boot = if (bootvar) boot_coefs else NULL,
    surv_curve = surv_curve
  ))
}

# TSRIF Cox model for time-varying effect

#' @noRd
est_tsrif_tvar <- function(formula, trtformula, data, trt, iv, fdist = "gamma",
                           t, bootvar = FALSE, B = NULL) {
  # Ensure the necessary libraries are loaded
  requireNamespace("survival")
  requireNamespace("stats")

  # Extract survival data
  data_outcome_model <- model.frame(formula, data = data)
  if (!inherits(data_outcome_model[,1], "Surv")) stop("The left-hand side of the outcome formula must be a 'Surv' object.")
  survtimes <- data_outcome_model[, 1][, 1]  # Extract survival times
  eventind <- data_outcome_model[, 1][, 2]  # Extract event indicators

  # Estimate survival function
  surv_curve <- survfit(Surv(survtimes, eventind) ~ 1)

  # First-stage model (using the original data)
  trt_model <- lm(trtformula, data = data)
  Resid <- as.vector(residuals(trt_model))
  data$Resid <- Resid

  # Split data for time-varying analysis
  data_long <- survSplit(formula = update(formula, . ~ . + Resid), data = data, cut = t, episode = "timegroup",
                         start = "tstart", end = "tstop", event = "event", id = "new_id")

  # Indices for each time period
  i1 <- which(data_long$timegroup == 1)
  i2 <- which(data_long$timegroup == 2)

  # Fit second-stage Cox models for first time period
  out_model1 <- coxph(
    #formula = update(formula, Surv(tstart, tstop, event) ~ . + Resid + frailty(1:length(i1), dist = fdist)),
    formula = as.formula(paste0(
      "Surv(tstart, tstop, event) ~ ",
      as.character(formula[3]),
      " + Resid + frailty(1:", length(i1), ", dist = '", fdist, "')"
    )),
    data = data_long[i1, ]
  )

  frail1 <- out_model1$frail
  Resid2 <- out_model1$coef["Resid"]^(-1) * frail1 + Resid
  data2 = as.data.frame(cbind(data, Resid2, frail1))
  data2_long= survSplit(formula = update(formula, . ~ . + Resid2 + frail1), data = data2, cut=t, episode ="timegroup",
                        start = "tstart", end = "tstop", event = "event", id = "new_id")

  # Fit the second Cox model for the second period, using updated residual
  out_model2 <- coxph(
    #formula = update(formula, Surv(tstart, tstop, event) ~ . + Resid2 + frailty(1:length(i2), dist = fdist)),
    formula = as.formula(paste0(
      "Surv(tstart, tstop, event) ~ ",
      as.character(formula[3]),
      " + Resid2 + frailty(1:", length(i2), ", dist = '", fdist, "')"
    )),
    data = data2_long[i2, ]
  )

  # Extract coefficients and variances
  est <- c(coef(out_model1)[trt], coef(out_model2)[trt])
  names(est) <- c(paste(trt,".t1",sep = ""), paste(trt,".t2",sep = ""))
  variance <- c(vcov(out_model1)[trt, trt], vcov(out_model2)[trt, trt])
  names(variance) <- c(paste(trt,".t1",sep = ""), paste(trt,".t2",sep = ""))

  if (bootvar) {
    # Bootstrap variance estimation
    N <- nrow(data)
    boot_coefs <- matrix(NA, nrow = B, ncol = 2)
    message("... Bootstrap simulation starts ...")

    for (b in 1:B) {
      # Bootstrap resampling
      boot_indices <- sample(1:N, replace = TRUE)
      boot_data <- data[boot_indices, ]

      # First-stage model on bootstrap sample
      boot_trt_model <- lm(trtformula, data = boot_data)
      boot_Resid <- residuals(boot_trt_model)
      boot_data$Resid <- boot_Resid

      # Split data into time-varying format
      boot_data_long <- survSplit(
        formula = update(formula, . ~ . + Resid), data = boot_data,
        cut = t, episode = "timegroup", start = "tstart", end = "tstop",
        event = "event", id = "new_id")

      # Indices for each time period
      i1 <- which(boot_data_long$timegroup == 1)
      i2 <- which(boot_data_long$timegroup == 2)

      # Fit the first period Cox model
      boot_model1 <- coxph(
        formula = as.formula(paste0(
          "Surv(tstart, tstop, event) ~ ",
          as.character(formula[3]),
          " + Resid + frailty(1:", length(i1), ", dist = '", fdist, "')"
        )),
        data = boot_data_long[i1, ]
      )
      boot_frail1 <- boot_model1$frail
      boot_Resid2 <- boot_model1$coef["Resid"]^(-1) * boot_frail1 + boot_Resid
      boot_data2 <- as.data.frame(cbind(boot_data, Resid2 = boot_Resid2, frail1 = boot_frail1))

      # Update the split data for the second period
      boot_data2_long <- survSplit(
        formula = update(formula, . ~ . + Resid2 + frail1), data = boot_data2,
        cut = t, episode = "timegroup", start = "tstart", end = "tstop",
        event = "event", id = "new_id")

      # Fit the second period Cox model
      boot_model2 <- coxph(
        formula = as.formula(paste0(
          "Surv(tstart, tstop, event) ~ ",
          as.character(formula[3]),
          " + Resid2 + frailty(1:", length(i2), ", dist = '", fdist, "')"
        )),
        data = boot_data2_long[i2, ]
      )

      # Store treatment effect estimates for each period
      boot_coefs[b, ] <- c(coef(boot_model1)[trt], coef(boot_model2)[trt])
    }
    message("... Bootstrap simulation ends ...")

    colnames(boot_coefs) <- c(paste(trt,".t1",sep = ""), paste(trt,".t2",sep = ""))

    # Calculate variances of bootstrap estimates
    variance <- apply(boot_coefs, 2, var)
  }

  # Organize some outputs
  cntrlfunc = as.data.frame(cbind(Resid, Resid2))

  # 95% Confidence Interval for treatment effect
  upper1 <- est[1] + qnorm(0.975) * sqrt(variance[1])
  lower1 <- est[1] - qnorm(0.975) * sqrt(variance[1])
  conf_int1 <- c(lower1, upper1)
  upper2 <- est[2] + qnorm(0.975) * sqrt(variance[2])
  lower2 <- est[2] - qnorm(0.975) * sqrt(variance[2])
  conf_int2 <- c(lower2, upper2)
  conf_int <- matrix(c(conf_int1, conf_int2), nrow = 2, byrow = TRUE)
  rownames(conf_int) <- names(est)
  colnames(conf_int) <- c("2.5%", "97.5%")

  trt_coef <- data.frame(
    logHR = unname(est),
    se = sqrt(variance),
    Z = unname(est) / sqrt(variance),
    p.val = 2 * pnorm(abs(unname(est) / sqrt(variance)), lower.tail = FALSE),
    row.names = names(est),
    check.names = FALSE
  )

  # Return results
  return(list(
    trt_coef = trt_coef,
    conf_int = conf_int,
    trt_model = trt_model,
    ctrl_func = cntrlfunc,
    out_model1 = out_model1,
    out_model2 = out_model2,
    est_boot = if (bootvar) boot_coefs else NULL,
    data_long = data2_long,
    surv_curve = surv_curve
  ))
}

#' Sequential trial emulation for time-dependent treatment data
#'
#' Constructs a stacked sequence of emulated trials from
#' longitudinal survival data in start-stop form, following the
#' sequential trials framework of Gran et al. and Keogh et al.
#'
#' Trial start times are determined only by the earliest observed cohort
#' entry time and by times at which at least one individual newly initiates
#' treatment (0 -> 1). Changes in time-varying covariates listed in
#' `covs` do **not** themselves create new trial start times. However,
#' `seqem()` still carries those covariates forward correctly into each
#' emulated trial by creating trial-specific baseline versions with suffix
#' `.base`.
#'
#' When `coarsen != "none"`, the observed timeline is discretized/coarsened onto a
#' grid of time width `cbin_width`. This coarsening is applied to **all** row
#' start times, not only to treatment changes, so that updates in
#' time-varying covariates are also mapped onto the coarsened scale.
#' With `coarsen = "floor"`, updates are mapped down to the beginning of
#' the containing interval; with `coarsen = "ceiling"`, they are mapped up
#' to the end of the containing interval. After coarsening, emulated trial
#' start times remain tied only to treatment initiation times on the
#' coarsened scale.
#'
#' The function algorithmic workflow:
#' \enumerate{
#'   \item Identifies trial start times: the earliest cohort start time plus
#'         all times at which any individual initiates treatment (`tvtrt =` 0 -> 1).
#'   \item Augments each individual's trajectory by splitting intervals at
#'         these trial start times so that each at-risk subject has an
#'         explicit row starting at each trial time.
#'   \item For each trial, includes individuals who are under observation at
#'         the trial start and have not yet started treatment, and constructs
#'         trial-specific baseline versions of treatment and covariates
#'         (with suffix `.base`), new follow-up times `start.new`, `stop.new`, `event`.
#'   \item Implements artificial censoring by restricting to rows where the
#'         current treatment equals the baseline treatment in that trial.
#' }
#'
#' @param data a `data.frame` in start-stop long format with potentially
#'   multiple rows per individual. Typically a dataset produced by
#'   [tmerge][survival::tmerge()].
#' @param start a string name of the interval start time variable.
#' @param stop a string name of the stop time variable.
#' @param event a string name of the event indicator (1: event).
#' @param tvtrt a string name of the time-varying **treatment** variable. Currently this
#'   must be binary and monotone in the sense assumed by the sequential
#'   trials framework (remain 1 once become 1). A trial-specific baseline
#'   version of `tvtrt` is always created as `tvtrt.base`.
#' @param id a string name of the unique individual identifier.
#' @param covs optional. a string vector of covariate names in `data` for
#'   which trial-specific baseline versions should be created. Both
#'   *time-constant* and *time-varying* covariates are supported. For each `X`
#'   in `covs`, `seqem()` creates `X.base`, representing the value of `X`
#'   at the start of each emulated trial for each individual.
#' @param coarsen one of `"none"`, `"floor"`, or `"ceiling"`.
#'   `"none"` leaves the original timeline unchanged. `"floor"` and
#'   `"ceiling"` discretize row start times to the coarsening grid and thus
#'   affect both treatment and time-varying covariate histories.
#' @param cbin_width a positive numeric bin width used when
#'   `coarsen = "floor"` or `"ceiling"`.
#'
#' @return A `seqem` object with components:
#' \itemize{
#'   \item `data_seqorig` : stacked emulated trials **before** artificial
#'         censoring, with columns:
#'         `id`, `trial`, `trial_time`, `rownum`, `visit`,
#'         original `start`, `stop`, `event`, `tvtrt`, covariates, and
#'         triap-specific follow-up intervals `start.new`, `stop.new`,
#'         `event`, trial-specific baseline treatment and covariates
#'         `{tvtrt}.base`, `{covs}.base` for covariates in `covs`.
#'   \item `data_seqem` : stacked emulated trials **after** artificial
#'         censoring, obtained by restricting to rows by forcing
#'         `{tvtrt} == {tvtrt}.base`. This is the dataset typically used for
#'         downstream `seqcox()` or `coxiv_seq()` with sequential trials
#'         emulation framework.
#'   \item `summary` : a character string describing the `seqem` object,
#'         used by `print.seqem()`.
#'   \item `trial_summary` : a per-trial summary `data.frame` with one row
#'         per emulated trial and columns:
#'         `trial` (trial index), `t0` (trial start time),
#'         `n` (unique individuals), `rows` (person-time rows),
#'         `treated` (unique baseline-treated individuals), and `events`
#'         (terminal events) in that trial.
#' }
#'
#' @examples
#' if (requireNamespace("survival", quietly = TRUE)) {
#'   data("heart", package = "survival")
#'
#'   fit_seqem <- seqem(data = heart, start = "start", stop = "stop",
#'     event = "event", tvtrt = "transplant", id = "id",
#'     covs = c("age", "surgery"), coarsen = "floor", cbin_width = 7)
#'
#'   print(fit_seqem)
#' }
#'
#' @references
#' 1. Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B, Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the causal effect of treatment in the presence of time-dependent confounding applied to data from the Swiss HIV Cohort Study. *Statistics in Medicine*. 2010 Nov 20;29(26):2757-68.
#' 2. Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in survival analysis using longitudinal observational data: Sequential trials and marginal structural models. *Statistics in Medicine*. 2023 Jun 15;42(13):2191-2225.
#'
#' @export

seqem <- function(data, start, stop, event, tvtrt, id, covs = NULL,
                  coarsen = c("none", "floor", "ceiling"), cbin_width = NULL) {

  coarsen <- match.arg(coarsen)

  ## ---- Basic checks ------------------------------------------------------

  if (!is.data.frame(data)) {
    stop("'data' must be a data.frame.")
  }
  for (arg in list(start = start, stop = stop, event = event,
                   tvtrt = tvtrt, id = id)) {
    nm  <- names(arg)
    val <- arg[[1L]]
    if (!is.character(val) || length(val) != 1L) {
      stop(sprintf("Argument '%s' must be a single character string.", nm))
    }
    if (!(val %in% names(data))) {
      stop(sprintf("Variable '%s' not found in 'data'.", val))
    }
  }

  if (!is.null(covs)) {
    if (!is.character(covs)) {
      stop("`covs` must be NULL or a character vector of variable names.")
    }
    missing_covs <- setdiff(covs, names(data))
    if (length(missing_covs) > 0L) {
      stop("The following covariates in `covs` are not in `data`: ",
           paste(missing_covs, collapse = ", "))
    }
  }

  if (coarsen != "none") {
    if (is.null(cbin_width) || !is.numeric(cbin_width) || length(cbin_width) != 1L || cbin_width <= 0) {
      stop("When `coarsen != \"none\"`, `cbin_width` must be a single positive numeric.")
    }
  }

  df <- data

  # Event as integer 0/1
  df[[event]] <- as.integer(df[[event]])

  # Sort by id and start time
  df <- df[order(df[[id]], df[[start]]), ]

  # At most one event per id
  ev_by_id <- tapply(df[[event]], df[[id]], function(x) sum(x, na.rm = TRUE))
  if (any(ev_by_id > 1L, na.rm = TRUE)) {
    stop("More than one event (", event, "==1) detected for some individuals. ",
         "The sequential trial emulation assumes a single terminal event per id.")
  }

  # tvtrt to numeric 0/1 if logical/factor
  if (is.logical(df[[tvtrt]])) {
    df[[tvtrt]] <- as.integer(df[[tvtrt]])
  } else if (is.factor(df[[tvtrt]])) {
    lev <- levels(df[[tvtrt]])
    if (length(lev) != 2L) {
      stop("Time-varying treatment '", tvtrt, "' is a factor with more than two levels. ",
           "Currently only binary treatment (0/1) is supported.")
    }
    df[[tvtrt]] <- as.integer(df[[tvtrt]] == lev[2L])
  }

  # Positive-length intervals
  if (any(df[[stop]] <= df[[start]])) {
    stop("Found rows with 'stop' <= 'start'. Ensure intervals are of positive length.")
  }

  ## ---- 0. Optional discretization / time coarsening (NEW) ---------------

  if (coarsen != "none") {

    origin <- min(df[[start]], na.rm = TRUE)

    snap_floor <- function(t) origin + floor((t - origin) / cbin_width) * cbin_width
    snap_ceil  <- function(t) origin + ceiling((t - origin) / cbin_width) * cbin_width

    snap_start <- if (coarsen == "floor") snap_floor else snap_ceil

    .coarsen_one_id <- function(df_i) {
      df_i <- df_i[order(df_i[[start]]), , drop = FALSE]

      entry <- min(df_i[[start]], na.rm = TRUE)
      exit  <- max(df_i[[stop]],  na.rm = TRUE)

      # True event time (if any): the stop time of the event row
      has_event <- any(df_i[[event]] == 1L, na.rm = TRUE)
      t_event   <- if (has_event) min(df_i[[stop]][df_i[[event]] == 1L], na.rm = TRUE) else Inf

      # Snap ALL row start times (these correspond to *any* TD covariate changes)
      s_raw <- df_i[[start]]
      s_new <- snap_start(s_raw)

      # Never create time before entry; keep the first start at the true entry
      s_new[1L] <- entry
      if (coarsen == "floor") {
        s_new <- pmax(s_new, entry)
      }

      # If multiple original rows map to the same snapped start, keep the LAST one
      # (i.e., latest observed values within that bin before/at the snapped boundary).
      grp <- split(seq_len(nrow(df_i)), s_new)
      keep_idx <- vapply(grp, max, integer(1))
      keep_idx <- keep_idx[order(as.numeric(names(keep_idx)))]

      df_c  <- df_i[keep_idx, , drop = FALSE]
      s_c   <- s_new[keep_idx]

      # Rebuild stop times from the next snapped start; last stop is min(exit, t_event)
      stop_c <- c(s_c[-1L], min(exit, t_event))

      # Assign new times, wipe events then set event once if applicable
      df_c[[start]] <- s_c
      df_c[[stop]]  <- stop_c
      df_c[[event]] <- 0L

      # Drop non-positive intervals (can happen if last snapped start >= exit/t_event)
      ok <- df_c[[stop]] > df_c[[start]]
      df_c <- df_c[ok, , drop = FALSE]
      if (nrow(df_c) == 0L) return(NULL)

      # Place the event in the interval containing t_event (and truncate after it)
      if (is.finite(t_event)) {
        j <- which(df_c[[start]] < t_event & df_c[[stop]] >= t_event)[1L]
        if (!is.na(j)) {
          df_c[[stop]][j]  <- t_event
          df_c[[event]][j] <- 1L
          df_c <- df_c[seq_len(j), , drop = FALSE]
        } else {
          # If no interval contains it (edge cases), ignore; original data should avoid this.
        }
      }

      df_c
    }

    ids <- unique(df[[id]])
    out_list <- vector("list", length(ids))
    for (ii in seq_along(ids)) {
      this_id <- ids[ii]
      df_i <- df[df[[id]] == this_id, , drop = FALSE]
      out_list[[ii]] <- .coarsen_one_id(df_i)
    }
    df2 <- do.call(rbind, Filter(Negate(is.null), out_list))
    rownames(df2) <- NULL
    df2 <- df2[order(df2[[id]], df2[[start]]), , drop = FALSE]

    # Re-check positive-length after coarsening
    if (any(df2[[stop]] <= df2[[start]])) {
      stop("Coarsening produced rows with 'stop' <= 'start'. ",
           "This should not happen; please report this example.")
    }

    df <- df2
  }

  ## ---- 1. Trial start times (baseline + treatment initiations) ----------

  prev_trt <- ave(df[[tvtrt]], df[[id]],
                  FUN = function(z) c(0, utils::head(z, -1L)))
  switch_rows    <- (df[[tvtrt]] == 1L) & (prev_trt == 0L)
  trt_init_times <- df[[start]][switch_rows]

  min_start   <- min(df[[start]], na.rm = TRUE)
  trial_times <- sort(unique(c(min_start, trt_init_times)))
  K           <- length(trial_times)
  if (K < 1L) stop("No valid trial start times found.")

  ## ---- 2. Augment each individual's trajectory at trial_times ----------

  .split_subject <- function(df_id, start_name, stop_name, event_name, cuts_subject) {
    if (length(cuts_subject) == 0L) return(df_id)

    out     <- vector("list", length = 0L)
    idx_out <- 1L

    for (i in seq_len(nrow(df_id))) {
      row <- df_id[i, , drop = FALSE]
      s0  <- row[[start_name]]
      s1  <- row[[stop_name]]
      ev  <- row[[event_name]]

      cuts_row <- cuts_subject[cuts_subject > s0 & cuts_subject < s1]

      if (length(cuts_row) == 0L) {
        out[[idx_out]] <- row
        idx_out <- idx_out + 1L
      } else {
        pts   <- c(s0, cuts_row, s1)
        n_seg <- length(pts) - 1L
        for (j in seq_len(n_seg)) {
          new_row               <- row
          new_row[[start_name]] <- pts[j]
          new_row[[stop_name]]  <- pts[j + 1L]
          if (j < n_seg) {
            new_row[[event_name]] <- 0L
          } else {
            new_row[[event_name]] <- ev
          }
          out[[idx_out]] <- new_row
          idx_out <- idx_out + 1L
        }
      }
    }

    do.call(rbind, out)
  }

  ids         <- unique(df[[id]])
  df_aug_list <- vector("list", length(ids))
  names(df_aug_list) <- as.character(ids)

  for (ii in seq_along(ids)) {
    this_id <- ids[ii]
    df_i    <- df[df[[id]] == this_id, , drop = FALSE]
    df_i    <- df_i[order(df_i[[start]]), ]

    s_min <- min(df_i[[start]])
    s_max <- max(df_i[[stop]])

    cuts_i <- trial_times[trial_times > s_min & trial_times < s_max]

    df_aug_list[[ii]] <- .split_subject(df_i,
                                        start_name = start,
                                        stop_name  = stop,
                                        event_name = event,
                                        cuts_subject = cuts_i)
  }

  df_aug <- do.call(rbind, df_aug_list)
  rownames(df_aug) <- NULL

  df_aug <- df_aug[order(df_aug[[id]], df_aug[[start]]), ]
  df_aug$visit <- ave(rep(1L, nrow(df_aug)), df_aug[[id]], FUN = cumsum)

  ## ---- 3. Precompute entry, exit, first treatment times -----------------

  id_levels <- sort(unique(df_aug[[id]]))

  entry_time <- tapply(df_aug[[start]], df_aug[[id]], min)
  exit_time  <- tapply(df_aug[[stop]],  df_aug[[id]], max)

  first_trt_time <- tapply(df_aug[[start]][df_aug[[tvtrt]] == 1L],
                           df_aug[[id]][df_aug[[tvtrt]] == 1L],
                           min)

  entry_time_all <- setNames(entry_time[match(id_levels, names(entry_time))],
                             id_levels)
  exit_time_all  <- setNames(exit_time[match(id_levels, names(exit_time))],
                             id_levels)
  first_trt_all  <- setNames(rep(Inf, length(id_levels)), id_levels)
  if (length(first_trt_time)) {
    first_trt_all[names(first_trt_time)] <- first_trt_time
  }

  ## ---- 4. Construct sequence of emulated trials (uncensored) -----------

  dat_seq_list <- vector("list", K)
  covs_for_baseline <- unique(c(covs, tvtrt))

  for (k in seq_along(trial_times)) {
    t0          <- trial_times[k]
    trial_index <- k - 1L  # 0-based as in Gran/Keogh

    elig_ids <- id_levels[
      entry_time_all <= t0 &
        exit_time_all  > t0 &
        first_trt_all  >= t0
    ]
    if (!length(elig_ids)) next

    idx_k   <- df_aug[[id]] %in% elig_ids & df_aug[[start]] >= t0
    trial_df <- df_aug[idx_k, , drop = FALSE]
    if (!nrow(trial_df)) next

    trial_df$trial      <- trial_index
    trial_df$trial_time <- t0

    trial_df <- trial_df[order(trial_df[[id]], trial_df[[start]]), ]

    trial_df$rownum <- ave(seq_len(nrow(trial_df)),
                           interaction(trial_df[[id]], trial_df$trial, drop = TRUE),
                           FUN = seq_along)

    # Baseline versions of tvtrt and covariates at trial start.
    # Because trajectories were already split at every trial start time,
    # time-varying covariates in `covs` are re-evaluated dynamically for
    # each emulated trial and become `{cov}.base` variables.
    first_rows <- !duplicated(trial_df[[id]])
    for (v in covs_for_baseline) {
      if (!v %in% names(trial_df)) {
        stop("Baseline covariate '", v, "' not found in trial data. ",
             "Check that it exists in the original data.")
      }
      base_vals <- trial_df[[v]][first_rows]
      names(base_vals) <- as.character(trial_df[[id]][first_rows])
      v_base <- base_vals[as.character(trial_df[[id]])]

      trial_df[[paste0(v, ".base")]] <- v_base
    }

    trial_df$start.new <- trial_df[[start]] - t0
    trial_df$stop.new  <- trial_df[[stop]]  - t0

    dat_seq_list[[k]] <- trial_df
  }

  dat_seq_list <- Filter(Negate(is.null), dat_seq_list)
  if (length(dat_seq_list) == 0L) {
    stop("No eligible individuals for any emulated trial after applying ",
         "entry/exit and 'not previously treated' criteria.")
  }

  dat_seq_orig <- do.call(rbind, dat_seq_list)
  rownames(dat_seq_orig) <- NULL

  ## ---- 5. Artificial censoring (MANDATORY) -----------------------------

  tv_base_name <- paste0(tvtrt, ".base")
  if (!tv_base_name %in% names(dat_seq_orig)) {
    stop("Baseline treatment variable '", tv_base_name,
         "' not found in stacked trials.")
  }

  keep    <- dat_seq_orig[[tvtrt]] == dat_seq_orig[[tv_base_name]]
  dat_seq <- dat_seq_orig[keep, , drop = FALSE]

  ## ---- 6. Attach attributes & build summary ----------------------------

  attr(dat_seq_orig, "trial_times") <- trial_times
  attr(dat_seq_orig, "id_var")      <- id
  attr(dat_seq_orig, "start_var")   <- start
  attr(dat_seq_orig, "stop_var")    <- stop
  attr(dat_seq_orig, "event_var")   <- event
  attr(dat_seq_orig, "tvtrt_var")   <- tvtrt
  attr(dat_seq_orig, "coarsen")     <- coarsen
  attr(dat_seq_orig, "cbin_width")  <- cbin_width

  attr(dat_seq, "trial_times") <- trial_times
  attr(dat_seq, "id_var")      <- id
  attr(dat_seq, "start_var")   <- start
  attr(dat_seq, "stop_var")    <- stop
  attr(dat_seq, "event_var")   <- event
  attr(dat_seq, "tvtrt_var")   <- tvtrt
  attr(dat_seq, "coarsen")     <- coarsen
  attr(dat_seq, "cbin_width")  <- cbin_width

  ## ---- 6b. Trial-level summary dataset ---------------------------------

  trial_var <- "trial"
  trial_summary <- NULL

  if (trial_var %in% names(dat_seq)) {
    trial_times_attr <- attr(dat_seq, "trial_times")
    id_var_attr      <- attr(dat_seq, "id_var")
    event_var_attr   <- attr(dat_seq, "event_var")
    tvtrt_var_attr   <- attr(dat_seq, "tvtrt_var")

    trials_present <- sort(unique(dat_seq[[trial_var]]))

    if (!is.null(id_var_attr) &&
        !is.null(event_var_attr) &&
        id_var_attr %in% names(dat_seq) &&
        event_var_attr %in% names(dat_seq)) {

      ids_per_trial <- tapply(dat_seq[[id_var_attr]], dat_seq[[trial_var]],
                              function(z) length(unique(z)))
      rows_per_trial <- tapply(seq_len(nrow(dat_seq)), dat_seq[[trial_var]], length)
      events_per_trial <- tapply(dat_seq[[event_var_attr]], dat_seq[[trial_var]],
                                 function(z) sum(z, na.rm = TRUE))

      # treated: number of unique ids with baseline treatment = 1 in each trial
      tv_base_name_attr <- if (!is.null(tvtrt_var_attr)) paste0(tvtrt_var_attr, ".base") else NULL
      treated_raw <- NULL
      if (!is.null(tv_base_name_attr) && tv_base_name_attr %in% names(dat_seq)) {
        treated_raw <- tapply(
          dat_seq[[id_var_attr]][dat_seq[[tv_base_name_attr]] == 1],
          dat_seq[[trial_var]][dat_seq[[tv_base_name_attr]] == 1],
          function(z) length(unique(z))
        )
      }

      align_vec <- function(v) v[match(trials_present, as.numeric(names(v)))]

      ids_aligned    <- align_vec(ids_per_trial)
      rows_aligned   <- align_vec(rows_per_trial)
      events_aligned <- align_vec(events_per_trial)
      if (!is.null(treated_raw)) {
        treated_aligned <- align_vec(treated_raw)
        treated_aligned[is.na(treated_aligned)] <- 0L
      } else {
        treated_aligned <- rep(NA_integer_, length(trials_present))
      }

      t0s <- if (!is.null(trial_times_attr)) {
        trial_times_attr[pmin(trials_present + 1L, length(trial_times_attr))]
      } else {
        rep(NA_real_, length(trials_present))
      }

      trial_summary <- data.frame(
        trial   = trials_present,
        t0      = t0s,
        n       = as.numeric(ids_aligned),
        rows    = as.numeric(rows_aligned),
        treated = as.numeric(treated_aligned),
        events  = as.numeric(events_aligned)
      )
      rownames(trial_summary) <- NULL
    }
  }

  summary_text <- .seqem_summary_text(dat_seq, trial_summary = trial_summary)

  out <- list(
    data_seqorig   = dat_seq_orig,
    data_seqem     = dat_seq,
    summary        = summary_text,
    trial_summary  = trial_summary
  )
  class(out) <- "seqem"
  out
}


# Internal helper: build multi-line summary string for a seqem object
.seqem_summary_text <- function(dat, trial_summary = NULL, max_show = 6L) {
  trial_times <- attr(dat, "trial_times")
  id_var      <- attr(dat, "id_var")
  tvtrt_var   <- attr(dat, "tvtrt_var")
  event_var   <- attr(dat, "event_var")
  coarsen     <- attr(dat, "coarsen")
  cbin_width  <- attr(dat, "cbin_width")

  if (!("trial" %in% names(dat))) {
    return("Object lacks 'trial' variable; not a valid STE dataset.\n")
  }

  n_ids <- if (!is.null(id_var) && id_var %in% names(dat)) {
    length(unique(dat[[id_var]]))
  } else {
    NA_integer_
  }

  trials_present   <- sort(unique(dat[["trial"]]))
  n_trials_present <- length(trials_present)
  n_trials_total   <- if (!is.null(trial_times)) length(trial_times) else n_trials_present

  if (is.null(trial_summary) || !is.data.frame(trial_summary) || nrow(trial_summary) == 0L) {
    trial_summary <- NULL
    if (!is.null(id_var) && !is.null(event_var) && id_var %in% names(dat) && event_var %in% names(dat)) {
      ids_per_trial <- tapply(dat[[id_var]], dat[["trial"]], function(z) length(unique(z)))
      events_per_trial <- tapply(dat[[event_var]], dat[["trial"]], function(z) sum(z, na.rm = TRUE))
      tv_base_name <- if (!is.null(tvtrt_var)) paste0(tvtrt_var, ".base") else NULL
      treated_raw <- NULL
      if (!is.null(tv_base_name) && tv_base_name %in% names(dat)) {
        treated_raw <- tapply(
          dat[[id_var]][dat[[tv_base_name]] == 1],
          dat[["trial"]][dat[[tv_base_name]] == 1],
          function(z) length(unique(z))
        )
      }
      align_vec <- function(v) v[match(trials_present, as.numeric(names(v)))]
      ids_aligned    <- align_vec(ids_per_trial)
      events_aligned <- align_vec(events_per_trial)
      if (!is.null(treated_raw)) {
        treated_aligned <- align_vec(treated_raw)
        treated_aligned[is.na(treated_aligned)] <- 0L
      } else {
        treated_aligned <- rep(NA_integer_, length(trials_present))
      }
      t0s <- if (!is.null(trial_times)) {
        trial_times[pmin(trials_present + 1L, length(trial_times))]
      } else {
        rep(NA_real_, length(trials_present))
      }
      trial_summary <- data.frame(
        trial = trials_present,
        t0 = t0s,
        n = as.numeric(ids_aligned),
        treated = as.numeric(treated_aligned),
        events = as.numeric(events_aligned)
      )
    }
  }

  fmt_num <- function(x) {
    format(x, digits = 6, trim = TRUE, scientific = FALSE)
  }

  fmt_trial_times <- function(x) {
    if (is.null(x) || !length(x)) return("NA")
    paste(fmt_num(x), collapse = ", ")
  }

  title <- "Sequential trial emulation data object (seqem)"
  soft_div <- strrep("~",  nchar(title))
  hard_div <- strrep("-",  nchar(title))

  buf <- character()
  buf <- c(buf, title, "\n")
  buf <- c(buf, strrep("=", nchar(title)), "\n")

  if (!is.na(n_ids)) {
    buf <- c(buf, sprintf("Sample size : %d\n", n_ids))
  }

  trial_line <- sprintf("Trials emulated : %d", n_trials_present)
  if (!is.null(trial_times)) {
    trial_line <- paste0(
      trial_line,
      sprintf("  (%d start times [ trial_time ])",
              n_trials_total)
    )
  }
  buf <- c(buf, trial_line, "\n")

  if (is.null(coarsen) || identical(coarsen, "none")) {
    buf <- c(buf, "Coarsen method : none\n")
  } else {
    buf <- c(buf, sprintf("Coarsen method : %s (per %s time units)\n",
                          coarsen, format(cbin_width, trim = TRUE, scientific = FALSE)))
  }

  buf <- c(buf, soft_div, "\n")
  buf <- c(buf, "Core variables:\n")
  if (!is.null(tvtrt_var)) {
    buf <- c(buf, sprintf("  Treatment (original) : [ %s ]\n", tvtrt_var))
    buf <- c(buf, sprintf("  Trial treatment : [ %s\n", paste0(tvtrt_var, ".base ]")))
  }
  buf <- c(buf, "  Trial : integer index [ trial ] (0-based)\n")
  if (all(c("start.new", "stop.new") %in% names(dat))) {
    buf <- c(buf, "  Trial intervals : [ start.new ], [ stop.new ], [ event ]\n")
  }
  buf <- c(buf, soft_div, "\n")

  if (!is.null(trial_summary) && is.data.frame(trial_summary) && nrow(trial_summary) > 0L) {
    buf <- c(buf, "Per-trial summary (present trials):\n")
    show_n <- min(nrow(trial_summary), max_show)
    for (j in seq_len(show_n)) {
      t0_str <- if (is.na(trial_summary$t0[j])) "NA" else fmt_num(trial_summary$t0[j])
      buf <- c(buf,
               sprintf("  - trial %d (t0 = %s) : ids = %d, treated = %d, events = %d\n",
                       trial_summary$trial[j],
                       t0_str,
                       trial_summary$n[j],
                       trial_summary$treated[j],
                       trial_summary$events[j]))
    }
    if (nrow(trial_summary) > show_n) {
      extra <- nrow(trial_summary) - show_n
      buf <- c(buf,
               sprintf("  ...... (%d additional trial%s not shown)\n",
                       extra, if (extra == 1L) "" else "s"))
    }
  }

  buf <- c(buf, hard_div, "\n")
  buf <- c(buf, "A `seqem` object\n")

  paste0(buf, collapse = "")
}

#' Print the output from a `coxivomom` model fit.
#'
#' Provide succinct and organized output for a `coxivomom` model, such as
#' `coxiv_omom()`.
#'
#' @param x an object of class `"coxivomom"`, usually created by
#'   `coxiv_omom()`.
#' @param digits the number of decimal places to display for coefficient
#'   summaries and hazard ratios.
#' @param ... additional arguments (currently unused).
#'
#' @return The input object `x`, invisibly. Called for its side effect of printing a summary to the console.
#'
#' @examples
#' data("VitD", package = "surviv")
#' fit_omom <- surviv::coxiv_omom(
#'   formula = survival::Surv(time, death) ~ vitd + age,
#'   trt = "vitd",
#'   iv = "filaggrin",
#'   data = VitD
#' )
#' print(fit_omom)
#'
#' @export
print.coxivomom <- function(x, digits = 3L, ...) {
  if (!inherits(x, "coxivomom")) {
    stop("Input must be of class 'coxivomom'.")
  }

  .print_text <- function(obj, ...) {
    txt <- utils::capture.output(print(obj, ...))
    cat(paste0(txt, collapse = "\n"), "\n")
  }

  .fmt_noquote <- function(mat, digits) {
    mat <- as.matrix(mat)
    out <- matrix(
      formatC(mat, format = "f", digits = digits),
      nrow = nrow(mat),
      ncol = ncol(mat),
      dimnames = dimnames(mat)
    )
    noquote(out)
  }

  .standardize_coef_df <- function(obj) {
    coef_df <- obj$coef
    if (is.null(coef_df)) {
      coef_df <- obj$est_coef
    }
    if (is.null(coef_df)) {
      stop("Cannot find coefficient summary in `x$coef` or `x$est_coef`.")
    }

    coef_df <- as.data.frame(coef_df, check.names = FALSE)

    nm <- names(coef_df)
    map <- c(
      "Estimate" = "logHR",
      "Std.Error" = "se",
      "Z.value" = "Z",
      "P(>|z|)" = "pval",
      "logHR" = "logHR",
      "se" = "se",
      "Z" = "Z",
      "pval" = "pval"
    )
    nm2 <- unname(map[nm])
    names(coef_df) <- ifelse(is.na(nm2), nm, nm2)

    needed <- c("logHR", "se")
    if (!all(needed %in% names(coef_df))) {
      stop("Coefficient summary must contain `logHR` and `se` columns.")
    }

    coef_df$logHR <- as.numeric(coef_df$logHR)
    coef_df$se <- as.numeric(coef_df$se)
    if ("Z" %in% names(coef_df)) coef_df$Z <- as.numeric(coef_df$Z)
    if ("pval" %in% names(coef_df)) coef_df$pval <- as.numeric(coef_df$pval)

    coef_df
  }

  .standardize_confint <- function(obj, rn = NULL) {
    ci <- obj$conf_int
    if (is.null(ci)) {
      stop("Cannot find confidence intervals in `x$conf_int`.")
    }

    ci <- as.data.frame(ci, check.names = FALSE)
    if (ncol(ci) < 2L) {
      stop("`x$conf_int` must have two columns.")
    }
    ci <- ci[, 1:2, drop = FALSE]
    names(ci) <- c("2.5%", "97.5%")
    ci[[1]] <- as.numeric(ci[[1]])
    ci[[2]] <- as.numeric(ci[[2]])

    if (!is.null(rn) && is.null(rownames(ci)) && nrow(ci) == length(rn)) {
      rownames(ci) <- rn
    }
    ci
  }

  coef_df <- .standardize_coef_df(x)
  conf_df <- .standardize_confint(x, rn = rownames(coef_df))

  trt_name <- if (!is.null(x$input) && !is.null(x$input$trt)) x$input$trt else NULL
  if (is.null(trt_name) || !nzchar(trt_name)) {
    trt_name <- rownames(coef_df)[1]
  }
  trt_row <- if (!is.null(rownames(coef_df)) && trt_name %in% rownames(coef_df)) {
    trt_name
  } else {
    rownames(coef_df)[1]
  }

  trt_est <- coef_df[trt_row, "logHR"]
  trt_lcl <- if (!is.null(rownames(conf_df)) && trt_row %in% rownames(conf_df)) conf_df[trt_row, 1] else conf_df[1, 1]
  trt_ucl <- if (!is.null(rownames(conf_df)) && trt_row %in% rownames(conf_df)) conf_df[trt_row, 2] else conf_df[1, 2]

  coef_tab <- cbind(
    logHR = coef_df$logHR,
    se = coef_df$se,
    HR = exp(coef_df$logHR)
  )
  rownames(coef_tab) <- rownames(coef_df)
  coef_tab <- .fmt_noquote(coef_tab, digits = digits)

  n_obs <- if (!is.null(x$n)) {
    as.integer(x$n[1])
  } else if (!is.null(x$surv_curve$n)) {
    as.integer(x$surv_curve$n[1])
  } else {
    NA_integer_
  }

  n_events <- if (!is.null(x$events)) {
    as.integer(x$events[1])
  } else if (!is.null(x$surv_curve$n.event)) {
    as.integer(sum(x$surv_curve$n.event, na.rm = TRUE))
  } else {
    NA_integer_
  }

  cat("Orthogonality method-of-moment IV estimation of Cox model\n")
  cat(strrep("=", 57), "\n")

  cat("Call:\n")
  print(x$call)
  cat(strrep("-", 57), "\n")

  if (!is.na(n_obs)) cat(sprintf("Sample size: %d\n", n_obs))
  if (!is.na(n_events)) cat(sprintf("Number of events: %d\n", n_events))
  cat(strrep("-", 57), "\n")

  cat(sprintf("Marginal treatment effect (%s):\n", trt_name))
  cat(sprintf("- Estimate: %.*f [%.*f, %.*f]\n", digits, trt_est, digits, trt_lcl, digits, trt_ucl))
  cat(sprintf(
    "- Hazard ratio: %.*f [%.*f, %.*f]\n",
    digits, exp(trt_est), digits, exp(trt_lcl), digits, exp(trt_ucl)
  ))
  cat(strrep("-", 57), "\n")

  cat("Coefficients:\n")
  .print_text(coef_tab, right = TRUE)
  cat(strrep("-", 57), "\n")

  if (!is.null(x$iv_diag$f_stat)) {
    cat("Nested exclusion test for weak IV:\n")
    cat(sprintf(
      "- F-statistics: %.*f%s\n",
      2, as.numeric(x$iv_diag$f_stat),
      ifelse(as.numeric(x$iv_diag$f_stat) > 10, " [> 10.0]", " [< 10.0]")
    ))
    cat(strrep("-", 57), "\n")
  }

  cat("A `coxivomom` object\n")
  invisible(x)
}

#' Orthogonality method-of-moment IV estimation of Cox model
#'
#' Estimate marginal causal treatment effect (in terms of log-hazard ratio) in Cox model setting subject to unmeasured confounding \eqn{U} using instrumental variable based on the IV **O**rthogonality **M**ethod-**o**f-**M**oment estimator in MacKenzie et al. (2014).
#'
#'
#' A causal Cox counterfactual survival model for survival time \eqn{T} with additive unmeasured confounding is assumed for treatment \eqn{A}, observed confounder \eqn{X}, and unmeasured confounder \eqn{U}.
#' \deqn{P(T(a)=t|T(a)\ge t, U=u) = \lambda_0(t)\exp(\beta_a a + \beta'x) + h(u,t)}
#' OMOM IV estimator assumes the following conditions: 1) relevance \eqn{Z \not\amalg A | U,X}, 2) exclusion restriction \eqn{Z\amalg (T,D)|A, U,X}, and 3) randomization \eqn{Z\amalg U|X}.
#' The estimates are obtained as the numerical solution to zero-crossing of estimating equation:
#' \deqn{\psi(\beta) \approx \sum^n_{i=1}\int^\tau_0 \left\{ Z_i-\frac{\sum^n_{j=1} Z_j D_j(s)\exp\{\beta_a A_j+\beta'X\}}{\sum^n_{j=1}D_j(s)\exp\{\beta_a A_j+\beta'X\}}dN_i(s) \right\} = 0 }
#' which is identical to the partial Cox score function for instrument \eqn{Z}. \eqn{D(t)} is the event indicator at time t and \eqn{N(t)} is a counting process for survival.
#' Huber-White sandwich estimator is used to obtain variance for \eqn{\hat{\beta}}.
#' Suggestive diagnostic test for weak IV was provided by fitting a pseudo-first stage treatment linear model (since OMOM IV is an estimating equation based IV approach, not two-stage based)
#' regressing treatment `trt` on IV `iv` and confounders extracted from all predictors from `formula` input except for `trt`.
#' A nested \link[stats]{anova} F-test was performed to compare fitted treatment models with and without `iv` as predictor.
#' A popular rule of thumb of indication for weak IV in econometric literature is when F-statistics is less than 10 from comparing IV-exclusion model.
#'
#'
#' @param formula a regression \link[stats]{formula} object with \link[survival]{Surv} object as outcome including survival time and censoring indicator.
#' @param trt a string indicates the name of the treatment variable of interest. Only time-invariant __binary__ or __continuous__ treatment is supported. It is strongly recommended that binary treatment is converted to numeric type (0,1) instead of using factor type. Treatment cannot be multi-categorical.
#' @param iv a string indicates the name of the instrumental variable. Only one time-variant __binary__ or __continuous__ instrumental variable is supported. IV cannot be multi-categorical.
#' @param data a \link[base]{data.frame} object of input data containing all variables specified in `formula`, including observed confounders.
#'
#' @return A `coxivomom` object for OMOM IV estimation of a Cox model with the following components:
#' * `call`: the match call of the OMOM IV model.
#' * `coef` : a `data.frame` of estimated log-hazard ratios and inference for all predictors in the outcome Cox model, with columns `logHR`, `se`, `Z`, and `p.val`.
#' * `conf_int` : a `data.frame` of 95% confidence interval(s) for the estimated log hazard ratios.
#' * `input`: the collected input of the OMOM IV model.
#' * `formula`: the outcome regression formula for Cox model
#' * `n`: the effective sample size used for model estimation after missingness removal.
#' * `events`: the effective number of events of survival outcome after missingness removal.
#' * `vcov_mat` : estimated covariance matrix of log hazard ratios.
#' * `iv_diag`: a list of IV diagnosis for weak IV. Include F-statistics and `anova` fit of nested treatment models comparison.
#' * `surv_curve`: a \link[survival]{survfit} object for fitted survival curve using complete data.
#'
#' @import MASS
#'
#' @examples
#' data("VitD", package = "surviv")
#' fit_omom <- coxiv_omom(survival::Surv(time, death) ~ vitd + age,
#'                         trt = "vitd", iv = "filaggrin", data = VitD)
#' fit_omom$coef
#' fit_omom$conf_int
#'
#' @references
#' MacKenzie, T.A., Tosteson, T.D., Morden, N.E. et al. Using instrumental variables to estimate a Cox proportional hazards regression subject to additive confounding. *Health Serv Outcomes Res Method* 14, 54-68 (2014).
#'
#' @export

coxiv_omom <- function(formula, trt, iv, data) {

  # Save input arguments
  call <- match.call()
  input <- as.list(environment())

  # Validate inputs
  if (missing(formula) || missing(trt) || missing(iv) || missing(data)) {
    stop("All arguments 'formula', 'trt', 'iv', and 'data' must be provided.")
  }

  if (!is.data.frame(data)) {
    stop("'data' must be a data frame.")
  }

  if (iv %in% attr(terms(formula), "term.labels")) {
    stop(paste("Instrumental variable '", iv, "' should not be included in outcome model as predictor.", sep = ""))
  }

  # Extract model frame
  data_outcome_model <- model.frame(formula, data = data)

  # Validate survival data
  if (!inherits(data_outcome_model[, 1], "Surv")) {
    stop("The left-hand side of the formula must be a 'Surv' object.")
  }

  # Extract time, censoring, and covariates
  survtimes <- data_outcome_model[, 1][, 1]
  eventind <- data_outcome_model[, 1][, 2]
  covariates <- data_outcome_model[, setdiff(colnames(data_outcome_model[,-1]), trt),
                                   drop = FALSE]

  # Validate treatment and instrumental variable
  if (!(trt %in% colnames(data))) {
    stop(paste("Treatment variable", trt, "is not found in the data."))
  }
  if (!(iv %in% colnames(data))) {
    stop(paste("Instrumental variable", iv, "is not found in the data."))
  }

  trt_values <- data[[trt]]


  iv_values <- data[[iv]]


  # Handle factors that are binary (two levels only)
  # Process passable treatment/IV cases with factor/string type
  if (is.factor(trt_values) && length(levels(trt_values)) == 2) {
    message("Attention! Binary factor data type detected in treatment.
            Converting to 0/1 based on factor levels.")
    trt_values <- as.numeric(trt_values) - 1
  } else if (is.character(trt_values) && length(unique(trt_values)) == 2) {
    message("Attention! Binary string data type detected in treatment.
            Converting to 0/1 based on factor levels")
    trt_values <- as.numeric(as.factor(trt_values)) -1
  }
  if (is.factor(iv_values) && length(levels(iv_values)) == 2) {
    message("Attention! Binary factor data type detected in instrumental variable.
            Converting to 0/1 based on factor levels.")
    iv_values <- as.numeric(iv_values) - 1
  } else if (is.character(iv_values) && length(unique(iv_values)) == 2) {
    message("Attention! Binary string data type detected in instrumental variable.
            Converting to 0/1 based on factor levels")
    iv_values <- as.numeric(as.factor(iv_values)) -1
  }

  # Ensure treatment and IV at thie point are numeric
  if (!is.numeric(trt_values)) {
    stop("The treatment variable must be continuous or binary (0/1).")
  }
  if (!is.numeric(iv_values)) {
    stop("The instrumental variable must be continuous or binary (0/1).")
  }

  # Check treatment and IV types
  if (all(trt_values %in% c(0, 1))) {
    message("~ Treatment variable is binary.")
  } else if (is.numeric(trt_values)) {
    message("~ Treatment variable is continuous.")
  } else {
    stop("The treatment variable must be binary or continuous.")
  }

  if (all(iv_values %in% c(0, 1))) {
    message("~ IV is binary.")
  } else if (is.numeric(iv_values)) {
    message("~ IV is continuous.")
  } else {
    stop("The treatment variable must be binary or continuous.")
  }

  # Process covariates
  for (col in colnames(covariates)) {
    if (is.character(covariates[[col]])) {
      # Convert character columns to factors
      covariates[[col]] <- as.factor(covariates[[col]])
    } else if (is.factor(covariates[[col]])) {
      # Keep factors as-is
      next
    } else if (!is.numeric(covariates[[col]])) {
      stop(paste("Unsupported data type for covariate:", col))
    }
  }

  # Convert covariates into a design matrix
  # model.matrix automatically handles factors (multi-level or binary)
  cov_matrix <- model.matrix(~ ., data = covariates)[, -1, drop = FALSE]

  # Handle missing data
  complete_data <- data.frame(
    survtimes = survtimes,
    eventind = eventind,
    trt_values = trt_values,
    iv_values = iv_values,
    cov_matrix
  )
  complete_data <- na.omit(complete_data)

  # Reassign variables after removing missing data
  survtimes <- complete_data$survtimes
  eventind <- complete_data$eventind
  trt_values <- complete_data$trt_values
  iv_values <- complete_data$iv_values
  covariates <- complete_data[, -(1:4), drop = FALSE]  # Remaining covariates

  # Reorder data by survival times
  ord <- order(survtimes, decreasing = TRUE)
  survtimes <- survtimes[ord]
  eventind <- eventind[ord]
  trt_values <- trt_values[ord]
  iv_values <- iv_values[ord]
  covariates <- covariates[ord, , drop = FALSE]

  # Combine treatment and covariates
  trt_cov <- as.matrix(cbind(trt_values, covariates))
  W <- as.matrix(cbind(iv_values, covariates))  # Combine IV and covariates

  # Initialize output objects
  S.W1 <- matrix(0, nrow = length(survtimes), ncol = ncol(W))

  # Estimation equation
  Est.Equat <- function(beta) {
    HR <- exp(trt_cov %*% beta)
    S.0 <- cumsum(HR)
    for (i in 1:ncol(W)) {
      S.W1[, i] <- cumsum(W[, i] * HR)
    }
    colSums(eventind * (W - S.W1 / S.0))
  }

  # Solve for beta using root-finding
  result <- tryCatch({
    rootSolve::multiroot(Est.Equat, start = rep(0, ncol(W)))
  }, error = function(e) {
    stop("Root finding algorithm failed: ", e$message)
  })

  # Handle cases where no root is found
  if (any(is.na(result$root))) {
    stop("Root finding algorithm failed: no solution found.")
  }

  beta_hat <- result$root

  # Variance estimation  (MacKenzie et al. OMOM sandwich)
  HR <- exp(trt_cov %*% beta_hat)
  S.0 <- cumsum(HR)

  # Recompute these at beta_hat (do NOT rely on S.W1 left over from multiroot)
  S.W1 <- matrix(0, nrow = length(survtimes), ncol = ncol(W))
  S.X1 <- matrix(0, nrow = length(survtimes), ncol = ncol(trt_cov))

  for (i in 1:ncol(W)) {
    S.W1[, i] <- cumsum(W[, i] * HR)
  }
  for (j in 1:ncol(trt_cov)) {
    S.X1[, j] <- cumsum(trt_cov[, j] * HR)
  }

  # Cross-cumsums needed for the Jacobian ("bread")
  S.W1X1 <- array(0, dim = c(length(survtimes), ncol(W), ncol(trt_cov)))
  for (i in 1:ncol(W)) {
    for (j in 1:ncol(trt_cov)) {
      S.W1X1[, i, j] <- cumsum(W[, i] * trt_cov[, j] * HR)
    }
  }

  # "Meat" and "bread"
  Var.E.E <- Deriv <- matrix(0, nrow = ncol(W), ncol = ncol(W))
  for (i in 1:ncol(W)) {
    for (j in 1:ncol(W)) {
      # A = sum dN * (W - E[W|risk]) (W - E[W|risk])'
      Var.E.E[i, j] <- sum(eventind *
                             (W[, i] - S.W1[, i] / S.0) *
                             (W[, j] - S.W1[, j] / S.0))
      Deriv[i, j] <- -sum(eventind *
                            (S.W1X1[, i, j] / S.0 -
                               (S.W1[, i] * S.X1[, j]) / (S.0^2)))
    }
  }

  Inv.Deriv <- tryCatch(
    solve(Deriv),
    error = function(e) MASS::ginv(Deriv)  # generalized inverse if singular
  )

  Sandwich <- Inv.Deriv %*% Var.E.E %*% t(Inv.Deriv)
  se <- sqrt(diag(Sandwich))

  # Add covariance matrix to output
  rownames(Sandwich) <- colnames(Sandwich) <- c(trt, colnames(cov_matrix))

  # Output results
  result_mat <- data.frame(
    logHR = beta_hat,
    se = se,
    Z = beta_hat / se,
    p.val = 2 * pnorm(abs(beta_hat / se), lower.tail = FALSE),
    row.names = c(trt, colnames(cov_matrix)),
    check.names = FALSE
  )

  # 95% Confidence Interval for treatment effect
  upper <- result_mat$logHR + qnorm(0.975) * result_mat$se
  lower <- result_mat$logHR - qnorm(0.975) * result_mat$se
  conf_int <- data.frame(lower, upper)
  rownames(conf_int) <- rownames(result_mat)
  names(conf_int) = c("2.5%", "97.5%")

  # Instrumental variable relevance diagnostics
  cders <- setdiff(attr(terms(formula), "term.labels"), trt)
  iv_diag <- iv_strgth(covs = cders, trt = trt, iv = iv, data = na.omit(data))
  if (iv_diag$f_stat < 10) message("Attention! Possible weak IV: nested anova has F < 10.0 for IV-exclusion treatment model.\n ")

  # Estimate survival function
  surv_curve <- survfit(Surv(survtimes, eventind) ~ 1, data = complete_data)

  # Organize output results
  result = list(
    call = call,
    input = input,
    formula = formula,
    n = nrow(complete_data),
    events = sum(eventind),
    coef = result_mat,
    vcov_mat = Sandwich,
    conf_int = conf_int,
    iv_diag = iv_diag,
    surv_curve = surv_curve
  )

  # Define classes
  class(result) <- c("coxivomom")

  result
}

#' @noRd

iv_strgth <- function(covs, trt, iv, data) {
  # Construct saturated and unsaturated treatment model formulas
  formula_sat <- as.formula(paste(trt, "~", paste(c(covs, iv), collapse = " + ")))
  if (is.null(covs)) {
    formula_unsat <- as.formula(paste(trt, "~", 1))
  } else {
    formula_unsat <- as.formula(paste(trt, "~", paste(covs, collapse = " + ")))
  }

  # Fit models
  trtmod_sat <- lm(formula_sat, data = data)
  trtmod_unsat <- lm(formula_unsat, data = data)

  # Perform ANOVA to compute F-statistic
  anova_res <- anova(trtmod_unsat, trtmod_sat)

  # Output IV diagnostics result
  res <- list(f_stat = anova_res$`F`[2],
              nested_anova = anova_res)

  return(res)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.seqcox.R
\name{print.seqcox}
\alias{print.seqcox}
\title{Print method for seqcox objects}
\usage{
\method{print}{seqcox}(x, digits = 4, ...)
}
\arguments{
\item{x}{An object of class \code{seqcox} returned by \code{\link[=seqcox]{seqcox()}}.}

\item{digits}{Number of decimal places for printed coefficient summaries.}

\item{...}{Ignored.}
}
\value{
The input object \code{x}, invisibly. Called for its side effect of printing a summary to the console.
}
\description{
Print method for seqcox objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.seqem.R
\name{print.seqem}
\alias{print.seqem}
\title{Print method for seqem objects}
\usage{
\method{print}{seqem}(x, ...)
}
\arguments{
\item{x}{An object of class \code{seqem} returned by \code{\link[=seqem]{seqem()}}.}

\item{...}{Ignored.}
}
\value{
The input object \code{x}, invisibly. Called for its side effect of printing a summary to the console.
}
\description{
Print method for seqem objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.coxivtsrif.R
\name{print.coxivtsrif}
\alias{print.coxivtsrif}
\title{Print the output from a \code{coxivtsrif} model fit.}
\usage{
\method{print}{coxivtsrif}(x, digits = 3L, ...)
}
\arguments{
\item{x}{an object of class \code{"coxivtsrif"}, usually created by
\code{coxiv_tsrif()}.}

\item{digits}{the number of decimal places to display for coefficient
summaries and hazard ratios.}

\item{...}{additional arguments (currently unused).}
}
\value{
The input object \code{x}, invisibly. Called for its side effect of printing a summary to the console.
}
\description{
Provide succinct and organized output for a \code{coxivtsrif} model, such as
\code{coxiv_tsrif()}.
}
\examples{
data("VitD", package = "surviv")
fit_tsrif <- surviv::coxiv_tsrif(
  surv = "time",
  cens = "death",
  covs = c("age"),
  trt = "vitd",
  iv = "filaggrin",
  tchange = NULL,
  data = VitD,
  fdist = "gaussian",
  bootvar = FALSE,
  B = 20,
  tvareff = FALSE
)
print(fit_tsrif)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VitDdata.R
\docType{data}
\name{VitD}
\alias{VitD}
\title{VitD: cohort data on vitamin D and mortality}
\format{
A data frame with variables:
\itemize{
\item \code{age} : Age at baseline
\item \code{filaggrin} : Indicator of filaggrin mutation
\item \code{vitd} : Serum 25-OH-D (nmol/L) at baseline
\item \code{time} : Follow-up time
\item \code{death} : Death indicator during follow-up
}
}
\source{
Mirrored from \pkg{ivtools} (LGPL (>= 3)).
}
\usage{
data(VitD)
}
\description{
This dataset is mirrored from the \pkg{ivtools} package (version 2.3.0).
The upstream documentation notes the data originate from a real cohort study
and were modified for public availability.
}
\references{
Sjolander, Arvid and Martinussen, Torben. "Instrumental Variable Estimation with the R Package ivtools" Epidemiologic Methods, vol. 8, no. 1, 2019, pp. 20180024.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coxiv_omom.R
\name{coxiv_omom}
\alias{coxiv_omom}
\title{Orthogonality method-of-moment IV estimation of Cox model}
\usage{
coxiv_omom(formula, trt, iv, data)
}
\arguments{
\item{formula}{a regression \link[stats]{formula} object with \link[survival]{Surv} object as outcome including survival time and censoring indicator.}

\item{trt}{a string indicates the name of the treatment variable of interest. Only time-invariant \strong{binary} or \strong{continuous} treatment is supported. It is strongly recommended that binary treatment is converted to numeric type (0,1) instead of using factor type. Treatment cannot be multi-categorical.}

\item{iv}{a string indicates the name of the instrumental variable. Only one time-variant \strong{binary} or \strong{continuous} instrumental variable is supported. IV cannot be multi-categorical.}

\item{data}{a \link[base]{data.frame} object of input data containing all variables specified in \code{formula}, including observed confounders.}
}
\value{
A \code{coxivomom} object for OMOM IV estimation of a Cox model with the following components:
\itemize{
\item \code{call}: the match call of the OMOM IV model.
\item \code{coef} : a \code{data.frame} of estimated log-hazard ratios and inference for all predictors in the outcome Cox model, with columns \code{logHR}, \code{se}, \code{Z}, and \code{p.val}.
\item \code{conf_int} : a \code{data.frame} of 95\% confidence interval(s) for the estimated log hazard ratios.
\item \code{input}: the collected input of the OMOM IV model.
\item \code{formula}: the outcome regression formula for Cox model
\item \code{n}: the effective sample size used for model estimation after missingness removal.
\item \code{events}: the effective number of events of survival outcome after missingness removal.
\item \code{vcov_mat} : estimated covariance matrix of log hazard ratios.
\item \code{iv_diag}: a list of IV diagnosis for weak IV. Include F-statistics and \code{anova} fit of nested treatment models comparison.
\item \code{surv_curve}: a \link[survival]{survfit} object for fitted survival curve using complete data.
}
}
\description{
Estimate marginal causal treatment effect (in terms of log-hazard ratio) in Cox model setting subject to unmeasured confounding \eqn{U} using instrumental variable based on the IV \strong{O}rthogonality \strong{M}ethod-\strong{o}f-\strong{M}oment estimator in MacKenzie et al. (2014).
}
\details{
A causal Cox counterfactual survival model for survival time \eqn{T} with additive unmeasured confounding is assumed for treatment \eqn{A}, observed confounder \eqn{X}, and unmeasured confounder \eqn{U}.
\deqn{P(T(a)=t|T(a)\ge t, U=u) = \lambda_0(t)\exp(\beta_a a + \beta'x) + h(u,t)}
OMOM IV estimator assumes the following conditions: 1) relevance \eqn{Z \not\amalg A | U,X}, 2) exclusion restriction \eqn{Z\amalg (T,D)|A, U,X}, and 3) randomization \eqn{Z\amalg U|X}.
The estimates are obtained as the numerical solution to zero-crossing of estimating equation:
\deqn{\psi(\beta) \approx \sum^n_{i=1}\int^\tau_0 \left\{ Z_i-\frac{\sum^n_{j=1} Z_j D_j(s)\exp\{\beta_a A_j+\beta'X\}}{\sum^n_{j=1}D_j(s)\exp\{\beta_a A_j+\beta'X\}}dN_i(s) \right\} = 0 }
which is identical to the partial Cox score function for instrument \eqn{Z}. \eqn{D(t)} is the event indicator at time t and \eqn{N(t)} is a counting process for survival.
Huber-White sandwich estimator is used to obtain variance for \eqn{\hat{\beta}}.
Suggestive diagnostic test for weak IV was provided by fitting a pseudo-first stage treatment linear model (since OMOM IV is an estimating equation based IV approach, not two-stage based)
regressing treatment \code{trt} on IV \code{iv} and confounders extracted from all predictors from \code{formula} input except for \code{trt}.
A nested \link[stats]{anova} F-test was performed to compare fitted treatment models with and without \code{iv} as predictor.
A popular rule of thumb of indication for weak IV in econometric literature is when F-statistics is less than 10 from comparing IV-exclusion model.
}
\examples{
data("VitD", package = "surviv")
fit_omom <- coxiv_omom(survival::Surv(time, death) ~ vitd + age,
                        trt = "vitd", iv = "filaggrin", data = VitD)
fit_omom$coef
fit_omom$conf_int

}
\references{
MacKenzie, T.A., Tosteson, T.D., Morden, N.E. et al. Using instrumental variables to estimate a Cox proportional hazards regression subject to additive confounding. \emph{Health Serv Outcomes Res Method} 14, 54-68 (2014).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coxiv_seq.R
\name{coxiv_seq}
\alias{coxiv_seq}
\title{Sequential 2SRI Cox analysis via sequential trials emulation}
\usage{
coxiv_seq(formula, trtformula, data, id, tvtrt, iv, stratify = FALSE,
  se_type = c("robust", "jackknife", "bootstrap", "none"), B = 50,
  seed = NULL, ipacw_den_covs = NULL, ipacw_num_covs = NULL,
  by_trial = TRUE, ...)
}
\arguments{
\item{formula}{a \code{formula} for the composite second-stage Cox model,
typically of the form
\code{Surv(start.new, stop.new, event) ~ tvtrt.base + x1.base + x2.base}.
The right-hand side must include \code{paste0(tvtrt, ".base")}. The control
function term \code{R.base} is added automatically if absent.}

\item{trtformula}{a \code{formula} for the trial-specific first-stage treatment
model. Its left-hand side must equal \code{paste0(tvtrt, ".base")}, and its
right-hand side must include \code{paste0(iv, ".base")}.}

\item{data}{a \code{seqem} object returned by \code{\link[=seqem]{seqem()}}.}

\item{id}{a string name for the subject identifier variable.}

\item{tvtrt}{a string name for the original time-varying treatment
variable.}

\item{iv}{a string name for the original baseline instrumental
variable.}

\item{stratify}{logical; if \code{TRUE}, the composite Cox model is stratified by
emulated trial so that each trial has its own baseline hazard.}

\item{se_type}{standard error estimator: one of \code{"robust"}, \code{"jackknife"},
\code{"bootstrap"}, or \code{"none"}.}

\item{B}{the number of bootstrap replicates when \code{se_type = "bootstrap"}.}

\item{seed}{optional random seed for bootstrap resampling.}

\item{ipacw_den_covs}{optional string vector of underlying covariate names
to use in the denominator IPACW model through their lead versions (for
example \code{c("endoleak")} uses \code{endoleak.lead1}). If \code{NULL}, the underlying
variables corresponding to non-treatment \code{.base} terms in \code{formula} and
\code{trtformula} are used whenever the original variables are available in the
stacked data.}

\item{ipacw_num_covs}{optional string vector of baseline \code{.base}
covariates to use in the numerator stabiliser model. If \code{NULL}, an
intercept-only numerator model is used.}

\item{by_trial}{logical; if \code{TRUE}, fit and return separate second-stage Cox
models within each emulated trial.}

\item{...}{additional arguments passed to \code{\link[survival:coxph]{survival::coxph()}}.}
}
\value{
A \code{coxivseq} object with components including:
\itemize{
\item \code{call} : the matched call.
\item \code{formula} : the composite second-stage Cox formula actually fitted.
\item \code{trtformula} : the sequential first-stage treatment models formula.
\item \code{fit} : the fitted composite \code{\link[survival:coxph]{survival::coxph()}} model.
\item \code{coef} : a \code{data.frame} with columns \code{logHR}, \code{HR}, \code{se}, \code{Z}, \code{p.val}.
\item \code{conf_int} : a \code{data.frame} of 95\% confidence limits with columns
\verb{2.5\%} and \verb{97.5\%} for estimated log hazard ratios.
\item \code{vcov_mat} : variance-covariance matrix corresponding to the
selected \code{se_type}.
\item \code{trtfit_by_trial} : list of first-stage treatment models.
\item \code{fit_by_trial} : list of second-stage trial-specific Cox fits when
\code{by_trial = TRUE}.
\item \code{data_ste} : final stacked, artificially censored analysis dataset,
including \code{R.base} and the computed IPACW column.
\item \code{jackknife}, \code{bootstrap}: a \code{matrix} of jackknife or bootstrap resampled
estimates if \code{se_type} set to \code{jackknife} or \code{boostrap}.
\item \code{tvtrt}, \code{iv}, \code{n}, \code{n_ids}, \code{n_trials}, \code{n_event}, \code{stratify},
\code{se_type}, \code{B}, \code{by_trial}: miscellaneous string tags for model print
summary.
}
}
\description{
Fits a sequential two-stage residual inclusion (2SRI) Cox
model for a time-varying treatment with a baseline instrumental variable,
using a stacked sequential trial emulation created by \code{\link[=seqem]{seqem()}}.
}
\details{
The analysis proceeds trial by trial. Within each emulated trial, a
first-stage logistic regression is fit for the trial-baseline treatment
\verb{\{tvtrt\}.base} on the baseline IV and baseline covariates from
\code{trtformula}. A trial-specific control-function residual \code{R.base} is then
constructed and carried across all rows within that subject-trial trajectory.

The stacked data are then artificially censored when observed treatment
deviates from the trial-baseline treatment status, and stabilised inverse
probability of artificial censoring weights (IPACW) are estimated. These
weights address the informative censoring induced by the sequential trial
emulation itself. This function does not estimate additional weights for
naturally occurring right censoring.

Finally, a weighted composite Cox model is fit on the artificially censored,
weighted stacked data. The returned coefficient summary is organised in the
same style as \code{\link[=seqcox]{seqcox()}}, \code{\link[=coxiv_omom]{coxiv_omom()}}, and \code{\link[=coxiv_tsrif]{coxiv_tsrif()}}.
}
\examples{
data("vascular", package = "surviv")

vasc_seqem <- seqem(
  data = vascular,
  start = "time.start",
  stop = "time.stop",
  event = "event",
  id = "id",
  tvtrt = "reint",
  covs = c("iv", "diameter", "endoleak"),
  coarsen = "ceiling",
  cbin_width = 2
)

fit_seqcox <- coxiv_seq(
  formula = survival::Surv(start.new, stop.new, event) ~
    reint.base + diameter.base + endoleak.base,
  trtformula = reint.base ~ iv.base + diameter.base + endoleak.base,
  data = vasc_seqem,
  id = "id",
  tvtrt = "reint",
  iv = "iv",
  se_type = "robust",
  by_trial = TRUE
)

print(fit_seqcox)

}
\references{
Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B,
Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the
causal effect of treatment in the presence of time-dependent confounding
applied to data from the Swiss HIV Cohort Study. \emph{Statistics in
Medicine}. 2010;29(26):2757-2768.

Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in
survival analysis using longitudinal observational data: Sequential trials
and marginal structural models. \emph{Statistics in Medicine}.
2023;42(13):2191-2225.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coxiv_tsrif.R
\name{coxiv_tsrif}
\alias{coxiv_tsrif}
\title{Two-stage residual inclusion-frailty (TSRI-F) instrumental variable analysis of Cox model}
\usage{
coxiv_tsrif(surv, cens, trt, iv, covs, data, tchange = NULL,
            tvareff = FALSE, fdist = "gamma", bootvar = FALSE, B = 50)
}
\arguments{
\item{surv}{a string indicating variable name of survival time \eqn{T}.}

\item{cens}{a string indicating variable name of event indicator \eqn{D}: 1 for terminal event occurrence and 0 for right censor.}

\item{trt}{a string indicating the treatment variable name \eqn{A}. Binary or continuous treatment supported.}

\item{iv}{a string or vector of strings indicating instrumental variable(s) \eqn{Z}. Binary or continuous IV supported.}

\item{covs}{a string or vector of strings indicating set of measured covariates \eqn{X}.}

\item{data}{a \link[base]{data.frame} containing all variables needed.}

\item{tchange}{a positive numeric value specifying the time point at which the treatment effect changes value.}

\item{tvareff}{logical; by default \code{FALSE} and estimate a time-constant treatment effect model. If \code{TRUE}, a time-varying treatment effect model is applied.}

\item{fdist}{the frailty distribution, default is \code{gamma}. Other options include \code{gaussian} or \code{t} distribution. Read more about frailty distribution specification at \link[survival]{frailty}.}

\item{bootvar}{logical; if \code{TRUE}, bootstrap variance estimation is performed. Otherwise, analytical variance is provided from frailty Cox model.}

\item{B}{number of bootstrap iterations if \code{bootvar} is \code{TRUE}.}
}
\value{
A \code{coxivtsrif} object with following components:
\itemize{
\item \code{input} : collected input and derived model formulas used for the TSRI-F fit.
\item \code{trt_coef} : a \code{data.frame} of treatment-effect estimates with columns \code{logHR}, \code{se}, \code{Z}, and \code{p.val}. If \code{tvareff = TRUE}, it contains one row for each treatment-effect period.
\item \code{conf_int} : a \code{data.frame} for 95\% confidence interval(s) for estimated treatment effect (s) from the choice of standard error estimator.
\item \code{ctrl_func} : the estimated control function \eqn{\widehat{R}}, or \code{data.frame} of \eqn{\widehat{R}^1} and \eqn{\widehat{R}^2} if \code{tvareff = TRUE}.
\item \code{trt_model} : a \link[stats]{lm} object fit of the first-stage treatment model.
\item \code{out_model} : a \link[survival]{coxph} object fit of the second-stage outcome Cox model. If \code{tvareff = TRUE}, Cox models for first period \code{out_model1} and second period \code{out_model2} are both returned.
\item \code{est_boot} :  a vector of length \code{B} of bootstrap estimates from each iteration. If \code{bootvar = FALSE}, returns \code{NULL}.
\item \code{data_long} :  a \code{data.frame} of data for model fit transformed into start-stop form for two-period Cox estimation, if \code{tvareff = TRUE}.
\item \code{surv_curve}: a \link[survival]{survfit} object for fitted survival curve using complete data.
\item \code{iv_diag} : a list of IV diagnostics for weak-IV assessment, including the nested ANOVA comparison and F-statistic.
}
}
\description{
Flexible instrumental analysis of Cox model in a novel \strong{T}wo-\strong{S}tage \strong{R}esidual \strong{I}nclusion with \strong{F}railty (TSRI-F) framework when treatment \eqn{A} and mortality \eqn{D} are subject to unmeasured confounding.
Allow estimation of time constant treatment effect (Martinez-Camblor et al. (2019)) and time-varying treatment effect (Martinez-Camblor et al. (2019))
in terms of log hazard ratio (log-HR). A set of instrumental variables \eqn{Z} and covariates for adjustment \eqn{X} (i.e., measured confounders) are required to estimate treatment effects \eqn{\beta^{(1)}_a} and \eqn{\beta^{(2)}_a} before and after the pre-specified time \code{tchange}.
}
\details{
This function performs two-stage residual inclusion IV analysis of the Cox model with individual frailty
when the treatment effect \eqn{\beta_a} is constant over time (\code{tvareff = FALSE}) or \eqn{\beta_a (t)} changes value at some time \emph{t} (\code{tvareff = TRUE}).
The IV assumptions are that 1) relevance \eqn{Z \not\amalg A | U,X}, 2) exclusion restriction \eqn{Z\amalg (T,D)|A, U,X}, and 3) randomization \eqn{Z\amalg U|X} with \eqn{U} the unmeasured confounder.
TSRI-F posits a first-stage linear model that generates the treatment: \deqn{A = \gamma_0+ \gamma_1Z+\gamma_2U+\gamma'X+\epsilon}
and a second-stage outcome model follows Cox proportional hazards form: \deqn{\lambda(t|X,U)=\lambda_0(t)\exp\{\beta_aA+\beta_uU+\beta'X\}}
where \eqn{\lambda_0} is the baseline hazard function. The first stage of TSRIF procedure for constant treatment effect estimates \emph{control function} \eqn{\widehat{R}} from residual of treatment model:
\eqn{\widehat{R} = A - \hat{\gamma}_0 -\hat{\gamma}_1Z-\hat{\gamma}'X}, and fit a second-stage Cox model with user-specified individual frailty \eqn{\phi} while adjusting for \eqn{\widehat{R}}.

For more intricate case of time-varying effect, the first stage proceeds identically to compute \eqn{\widehat{R}^{(1)} = A - \hat{\gamma}_0 -\hat{\gamma}_1Z-\hat{\gamma}'X}.
Second stage involves fitting two Cox models to estimate \eqn{\beta_a^{(1)} = \beta_a(t), t<} \code{tchange} and \eqn{\beta_a^{(2)} = \beta_a(t), t\ge} \code{tchange}.
The first-period Cox model is fitted with frailty term \eqn{\hat{\phi_1}} and including \eqn{\widehat{R}^{(1)}} as a covariate.
Then, update the control function \eqn{\widehat{R}^{(2)} = \widehat{R}^{(1)} + \hat{\beta}^{(1), -1}_r \hat{\phi_1}} for the second period, which are then included as a covariate in the second-period Cox model
with a new frailty term.

This approach estimates conditional treatment effect of \code{trt}. Coefficient estimates of covariates other than \code{trt} are also estimated and readily available, but these estimates are
suggestive of their associational relations with survival outcome, and causal conclusions regarding these covariates with outcome should be cautious.
}
\examples{
data("VitD", package = "surviv")

fit_tsrif1 <- coxiv_tsrif(surv = "time", cens = "death", covs = c("age"),
                          trt = "vitd", iv = "filaggrin", tchange = NULL,
                          data = VitD, fdist = "gaussian", bootvar = FALSE,
                          B = 20, tvareff = FALSE)
fit_tsrif1$trt_coef

fit_tsrif2 <- coxiv_tsrif(surv = "time", cens = "death", covs = c("age"),
                          trt = "vitd", iv = "filaggrin", tchange = 5,
                          data = VitD, fdist = "gaussian", bootvar = FALSE,
                          B = 20, tvareff = TRUE)
fit_tsrif2$trt_coef

}
\references{
\enumerate{
\item Pablo Martinez-Camblor, Todd Mackenzie, Douglas O Staiger, Philip P Goodney, A James O'Malley, Adjusting for bias introduced by instrumental variable estimation in the Cox proportional hazards model, \emph{Biostatistics}, Volume 20, Issue 1, January 2019, Pages 80-96,
\item Pablo Martinez-Camblor, Todd A. MacKenzie, Douglas O. Staiger, Phillip P. Goodney, A. James O'Malley, An Instrumental Variable Procedure for Estimating Cox Models with Non-Proportional Hazards in the Presence Of Unmeasured Confounding, \emph{Journal of the Royal Statistical Society Series C: Applied Statistics}, Volume 68, Issue 4, August 2019, Pages 985-1005
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.coxivomom.R
\name{print.coxivomom}
\alias{print.coxivomom}
\title{Print the output from a \code{coxivomom} model fit.}
\usage{
\method{print}{coxivomom}(x, digits = 3L, ...)
}
\arguments{
\item{x}{an object of class \code{"coxivomom"}, usually created by
\code{coxiv_omom()}.}

\item{digits}{the number of decimal places to display for coefficient
summaries and hazard ratios.}

\item{...}{additional arguments (currently unused).}
}
\value{
The input object \code{x}, invisibly. Called for its side effect of printing a summary to the console.
}
\description{
Provide succinct and organized output for a \code{coxivomom} model, such as
\code{coxiv_omom()}.
}
\examples{
data("VitD", package = "surviv")
fit_omom <- surviv::coxiv_omom(
  formula = survival::Surv(time, death) ~ vitd + age,
  trt = "vitd",
  iv = "filaggrin",
  data = VitD
)
print(fit_omom)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqem.R
\name{seqem}
\alias{seqem}
\title{Sequential trial emulation for time-dependent treatment data}
\usage{
seqem(data, start, stop, event, tvtrt, id, covs = NULL, coarsen = c("none",
  "floor", "ceiling"), cbin_width = NULL)
}
\arguments{
\item{data}{a \code{data.frame} in start-stop long format with potentially
multiple rows per individual. Typically a dataset produced by
\link[survival:tmerge]{tmerge}.}

\item{start}{a string name of the interval start time variable.}

\item{stop}{a string name of the stop time variable.}

\item{event}{a string name of the event indicator (1: event).}

\item{tvtrt}{a string name of the time-varying \strong{treatment} variable. Currently this
must be binary and monotone in the sense assumed by the sequential
trials framework (remain 1 once become 1). A trial-specific baseline
version of \code{tvtrt} is always created as \code{tvtrt.base}.}

\item{id}{a string name of the unique individual identifier.}

\item{covs}{optional. a string vector of covariate names in \code{data} for
which trial-specific baseline versions should be created. Both
\emph{time-constant} and \emph{time-varying} covariates are supported. For each \code{X}
in \code{covs}, \code{seqem()} creates \code{X.base}, representing the value of \code{X}
at the start of each emulated trial for each individual.}

\item{coarsen}{one of \code{"none"}, \code{"floor"}, or \code{"ceiling"}.
\code{"none"} leaves the original timeline unchanged. \code{"floor"} and
\code{"ceiling"} discretize row start times to the coarsening grid and thus
affect both treatment and time-varying covariate histories.}

\item{cbin_width}{a positive numeric bin width used when
\code{coarsen = "floor"} or \code{"ceiling"}.}
}
\value{
A \code{seqem} object with components:
\itemize{
\item \code{data_seqorig} : stacked emulated trials \strong{before} artificial
censoring, with columns:
\code{id}, \code{trial}, \code{trial_time}, \code{rownum}, \code{visit},
original \code{start}, \code{stop}, \code{event}, \code{tvtrt}, covariates, and
triap-specific follow-up intervals \code{start.new}, \code{stop.new},
\code{event}, trial-specific baseline treatment and covariates
\verb{\{tvtrt\}.base}, \verb{\{covs\}.base} for covariates in \code{covs}.
\item \code{data_seqem} : stacked emulated trials \strong{after} artificial
censoring, obtained by restricting to rows by forcing
\verb{\{tvtrt\} == \{tvtrt\}.base}. This is the dataset typically used for
downstream \code{seqcox()} or \code{coxiv_seq()} with sequential trials
emulation framework.
\item \code{summary} : a character string describing the \code{seqem} object,
used by \code{print.seqem()}.
\item \code{trial_summary} : a per-trial summary \code{data.frame} with one row
per emulated trial and columns:
\code{trial} (trial index), \code{t0} (trial start time),
\code{n} (unique individuals), \code{rows} (person-time rows),
\code{treated} (unique baseline-treated individuals), and \code{events}
(terminal events) in that trial.
}
}
\description{
Constructs a stacked sequence of emulated trials from
longitudinal survival data in start-stop form, following the
sequential trials framework of Gran et al. and Keogh et al.
}
\details{
Trial start times are determined only by the earliest observed cohort
entry time and by times at which at least one individual newly initiates
treatment (0 -> 1). Changes in time-varying covariates listed in
\code{covs} do \strong{not} themselves create new trial start times. However,
\code{seqem()} still carries those covariates forward correctly into each
emulated trial by creating trial-specific baseline versions with suffix
\code{.base}.

When \code{coarsen != "none"}, the observed timeline is discretized/coarsened onto a
grid of time width \code{cbin_width}. This coarsening is applied to \strong{all} row
start times, not only to treatment changes, so that updates in
time-varying covariates are also mapped onto the coarsened scale.
With \code{coarsen = "floor"}, updates are mapped down to the beginning of
the containing interval; with \code{coarsen = "ceiling"}, they are mapped up
to the end of the containing interval. After coarsening, emulated trial
start times remain tied only to treatment initiation times on the
coarsened scale.

The function algorithmic workflow:
\enumerate{
\item Identifies trial start times: the earliest cohort start time plus
all times at which any individual initiates treatment (\verb{tvtrt =} 0 -> 1).
\item Augments each individual's trajectory by splitting intervals at
these trial start times so that each at-risk subject has an
explicit row starting at each trial time.
\item For each trial, includes individuals who are under observation at
the trial start and have not yet started treatment, and constructs
trial-specific baseline versions of treatment and covariates
(with suffix \code{.base}), new follow-up times \code{start.new}, \code{stop.new}, \code{event}.
\item Implements artificial censoring by restricting to rows where the
current treatment equals the baseline treatment in that trial.
}
}
\examples{
if (requireNamespace("survival", quietly = TRUE)) {
  data("heart", package = "survival")

  fit_seqem <- seqem(data = heart, start = "start", stop = "stop",
    event = "event", tvtrt = "transplant", id = "id",
    covs = c("age", "surgery"), coarsen = "floor", cbin_width = 7)

  print(fit_seqem)
}

}
\references{
\enumerate{
\item Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B, Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the causal effect of treatment in the presence of time-dependent confounding applied to data from the Swiss HIV Cohort Study. \emph{Statistics in Medicine}. 2010 Nov 20;29(26):2757-68.
\item Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in survival analysis using longitudinal observational data: Sequential trials and marginal structural models. \emph{Statistics in Medicine}. 2023 Jun 15;42(13):2191-2225.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqcox.R
\name{seqcox}
\alias{seqcox}
\title{Sequential Cox analysis via sequential trials emulation}
\usage{
seqcox(formula, data, id, trial_id = "trial", stratify = FALSE,
  se_type = c("robust", "jackknife", "bootstrap", "none"), B = 50,
  seed = NULL, ipacw_trunc = 0.95, ipacw_den_covs = NULL,
  ipacw_num_covs = NULL, by_trial = TRUE, ...)
}
\arguments{
\item{formula}{a survival formula for the composite Cox model, typically of
the form \code{Surv(start.new, stop.new, event) ~ trt.base + L.base + ...}.
The \emph{first} \code{.base} term is treated as the trial-baseline treatment variable.}

\item{data}{a \code{seqem} object returned by \code{\link[=seqem]{seqem()}}. A raw data frame is not
accepted.}

\item{id}{a string name of the subject identifier column in the sequential trials data.}

\item{trial_id}{a string name of the trial index column. Defaults to \code{"trial"}.}

\item{stratify}{logical; if \code{TRUE}, the composite Cox model is stratified by
\code{trial_id} so each emulated trial has its own baseline hazard.}

\item{se_type}{standard error estimator: one of \code{"robust"}, \code{"jackknife"},
\code{"bootstrap"}, or \code{"none"}.}

\item{B}{number of bootstrap replicates when \code{se_type = "bootstrap"}.}

\item{seed}{optional random seed for the bootstrap.}

\item{ipacw_trunc}{optional truncation quantile for cumulative stabilized
IPACW. Defaults to \code{0.95}. Set to \code{NULL} to disable truncation.}

\item{ipacw_den_covs}{optional character vector of underlying covariate names
to use in the denominator IPACW model through their lead versions (for
example \code{c("endoleak")} uses \code{endoleak.lead1}). If \code{NULL}, the underlying
variables corresponding to non-treatment \code{.base} terms in \code{formula} are
used.}

\item{ipacw_num_covs}{optional character vector of baseline \code{.base}
covariates to use in the numerator stabilizer model. If \code{NULL}, an
intercept-only numerator model is used.}

\item{by_trial}{logical; if \code{TRUE} (default), fit trial-specific Cox models
and return them in \verb{$fit_by_trial}.}

\item{...}{additional arguments passed to \code{\link[survival:coxph]{survival::coxph()}}.}
}
\value{
A \code{seqcox} object with components including:
\itemize{
\item \code{call} : the matched call.
\item \code{formula} : the composite Cox formula actually fitted.
\item \code{fit} : a composite \code{\link[survival:coxph]{survival::coxph()}} fit.
\item \code{coef} : a \code{data.frame} of estimated coefficient summaries with columns
\code{logHR}, \code{HR}, \code{se}, \code{Z}, and \code{p.val}.
\item \code{conf_int} : a \code{data.frame} of 95\% confidence limits with columns
\verb{2.5\%} and \verb{97.5\%}.
\item \code{n_ids}, \code{n_trials}, \code{n_event}: the analysis size summaries
for sample size, number of emulated trials aggregated, and number
of events.
\item \code{fit_by_trial} : a list of trial-specific Cox fits when
\code{by_trial = TRUE}.
\item \code{data_ste} : a \code{data.frame} of stacked analysis dataset actually used in
the composite fit, including the computed IPACW column.
\item \code{jackknife}, \code{bootstrap}: a vector of jackknife or bootstrap resampled
estimates if \code{se_type} set to \code{jackknife} or \code{boostrap}.
}
}
\description{
Fit a composite Cox model to a stacked sequential trial emulation object
produced by \code{\link[=seqem]{seqem()}}. The analysis follows the sequential Cox approach of
Gran et al. by using stabilized inverse probability of artificial censoring
weights (IPACW) to adjust for the informative artificial censoring induced by
restricting each emulated trial to individuals who continue to follow their
treatment assignment at the trial start.
}
\examples{
  data("vascular", package = "surviv")
  vasc_seqem = seqem(data = vascular, start = "time.start",
                     stop = "time.stop", event = "event", id = "id",
                     tvtrt = "reint", covs = c("diameter", "endoleak"),
                     coarsen = "ceiling", cbin_width = 2)
  fit_seqcox <- seqcox(
    formula = survival::Surv(start.new, stop.new, event) ~
      reint.base + diameter.base + endoleak.base,
    data = vasc_seqem,
    id = "id",
    se_type = "robust"
  )
  print(fit_seqcox)

}
\references{
Gran JM, Roysland K, Wolbers M, Didelez V, Sterne JA, Ledergerber B,
Furrer H, von Wyl V, Aalen OO. A sequential Cox approach for estimating the
causal effect of treatment in the presence of time-dependent confounding
applied to data from the Swiss HIV Cohort Study. \emph{Statistics in
Medicine}. 2010;29(26):2757-2768.

Keogh RH, Gran JM, Seaman SR, Davies G, Vansteelandt S. Causal inference in
survival analysis using longitudinal observational data: Sequential trials
and marginal structural models. \emph{Statistics in Medicine}.
2023;42(13):2191-2225.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vasculardata.R
\docType{data}
\name{vascular}
\alias{vascular}
\title{vascular: longitudinal vascular surgery follow-up data}
\format{
A data frame with variables:
\itemize{
\item \code{id} : Unique patient identifier
\item \code{surv.time} : Observed overall follow-up time to death or censoring
\item \code{death} : Terminal event indicator
\item \code{time.start} : Interval start time
\item \code{time.stop} : Interval stop time
\item \code{event} : Counting-process event indicator for the interval
\item \code{reint} : Time-varying reintervention status at the current visit
\item \code{endoleak} : Time-varying endoleak status at the current visit
\item \code{iv} : Baseline instrumental variable for reintervention propensity
\item \code{smoke} : Baseline smoking status indicator
\item \code{age} : Baseline age
\item \code{diameter} : Baseline aneurysm diameter
\item \code{gender} : Baseline gender indicator
\item \code{hyper} : Baseline hypertension indicator
\item \code{obese} : Baseline obesity indicator
\item \code{reintlag1} : Lagged reintervention status from the previous interval
\item \code{visit} : Visit number
}
}
\source{
Synthetic data generated to mimic the distributional features of a
vascular surgery longitudinal follow-up study.
}
\usage{
data(vascular)
}
\description{
A synthetic longitudinal dataset designed to mimic repeated follow-up in a
vascular surgery study. The data are stored in long start-stop format for
survival analyses with a time-varying exposure and a time-varying binary
prognostic factor. In particular, \code{reint} indicates reintervention surgery
status over time and \code{endoleak} indicates time-varying endoleak status over
time. Both are coded so that once the status becomes 1 it remains 1 at later
visits.
}
\keyword{datasets}
